package sos.connection;


import java.util.Properties;

import java.sql.Driver;
import java.sql.Statement;
import java.sql.Connection;

import sos.util.SOSClassUtil;
import sos.util.SOSLogger;
import sos.util.SOSString;


/**
 * Title: 
 * <p>Description: Implementierung der SOSConnection-Klasse f�r MSSQLle</p>
 * Copyright: Copyright (c) 2003
 * Company: SOS GmbH
 * @author <a href="mailto:ghassan.beydoun@sos-berlin.com">Ghassan Beydoun</a>
 * @version $Id: SOSMSSQLConnection.java 937 2005-06-01 07:48:01Z gb $
 */


public class SOSMSSQLConnection
    extends sos.connection.SOSConnection {

  /** Umsetzung f�r %lcase, %ucase, %now, %updlock */
  private static final String replacement[] = {
      "LOWER", "UPPER", "GETDATE()", "WITH (UPDLOCK)"};



  /**
   *
   * @param connection
   * @param logger
   * @throws java.lang.Exception
   */
  public SOSMSSQLConnection ( Connection connection, SOSLogger logger) throws
      Exception {
    super(connection, logger);
    super.replacement = replacement;
    prepare( connection) ;
  }


  /**
  *
  * @param connection
  * @throws java.lang.Exception
  */
 public SOSMSSQLConnection ( Connection connection) throws
     Exception {
    super(connection);
   super.replacement = replacement;
   prepare( connection) ;
 }
  

  /**
   * Konstruktor f�r SOSMSSQLConnection
   *
   * @param configFileName
   * @param logger logging Objekt
   * @throws java.lang.Exception
   * @see #SOSMSSQLConnection(String)
   * @see #SOSMSSQLConnection(String, String, String, String, String SOSLogger)
   * @see #SOSMSSQLConnection(String, String, String, String, String)
   */
  public SOSMSSQLConnection(String configFileName, SOSLogger logger) throws
      Exception {

    super(configFileName, logger);
    super.replacement = replacement;
  }


  /**
   * Konstruktor f�r SOSMSSQLConnection
   *
   * @param configFileName
   * @throws java.lang.Exception
   * @see #SOSMSSQLConnection(String, SOSLogger)
   * @see #SOSMSSQLConnection(String, String, String, String, String, SOSLogger)
   * @see #SOSMSSQLConnection(String, String, String, String, String)
   */
  public SOSMSSQLConnection(String configFileName) throws
      Exception {

    super(configFileName);
    super.replacement = replacement;
  }


  /**
   * Konstruktor
   *
   * @param driver
   * @param url
   * @param dbuser
   * @param dbpassword
   * @throws java.lang.Exception
   * @see #SOSMySQLConnection(String)
   * @see #SOSMySQLConnection(String, SOSLogger)
   * @see #SOSMySQLConnection(String, String, String, String, SOSLogger)
   */
  public SOSMSSQLConnection(String driver, String url, String dbuser, String dbpassword) throws Exception {

    super( driver, url, dbuser, dbpassword);
    super.replacement = replacement;
  }


  
  /**
   * Konstruktor
   *
   * @param driver
   * @param url
   * @param dbuser
   * @param dbpassword
   * @throws java.lang.Exception
   * @see #SOSMySQLConnection(String)
   * @see #SOSMySQLConnection(String, SOSLogger)
   * @see #SOSMySQLConnection(String, String, String, String, SOSLogger)
   */
  public SOSMSSQLConnection(String driver, String url, String dbuser, String dbpassword, SOSLogger logger) throws Exception {

    super( driver, url, dbuser, dbpassword, logger);
    super.replacement = replacement;
  }


  
  /**
   * Konstruktor
   *
   * @param driver
   * @param url
   * @param dbname
   * @param dbuser
   * @param dbpassword
   * @param logger
   * @throws java.lang.Exception
   * @see #SOSMSSQLConnection(String)
   * @see #SOSMSSQLConnection(String, SOSLogger)
   * @see #SOSMSSQLConnection(String, String, String, String, String)
   */
  public SOSMSSQLConnection(String driver, String url, String dbname,
                            String dbuser, String dbpassword, SOSLogger logger) throws
      Exception {

    super(driver, url, dbuser, dbpassword, logger);
    super.replacement = replacement;
    if (dbname == null)
      throw new Exception(SOSClassUtil.getMethodName() +
                          ": missing database name.");
    this.dbname = dbname;
  }

  /**
   * Konstruktor
   *
   * @param driver
   * @param url
   * @param dbname
   * @param dbuser
   * @param dbpassword
   * @throws java.lang.Exception
   * @see #SOSMSSQLConnection(String)
   * @see #SOSMSSQLConnection(String, SOSLogger)
   * @see #SOSMSSQLConnection(String, String, String, String, String, SOSLogger)
   */
  public SOSMSSQLConnection(String driver, String url, String dbname,
                            String dbuser, String dbpassword) throws
      Exception {

    super(driver, url, dbuser, dbpassword);
    super.replacement = replacement;
    if (dbname == null)
      throw new Exception(SOSClassUtil.getMethodName() +
                          ": missing database name.");
    this.dbname = dbname;
  }


  /**
   * stellt die Datenbankverbindung her
   *
   * @throws java.lang.Exception falls ein Datenbankfehler vorliegt.
   */
  public void connect() throws Exception {

    Properties properties = new Properties();

    if ( SOSString.isEmpty(url))
      throw new Exception(SOSClassUtil.getMethodName() +
                          ": missing database url.");
    if ( SOSString.isEmpty(driver) )
      throw new Exception(SOSClassUtil.getMethodName() +
                          ": missing database driver.");
    if ( SOSString.isEmpty(dbuser))
      throw new Exception(SOSClassUtil.getMethodName() +
                          ": missing database user.");
    if ( SOSString.isEmpty(dbpassword))
      throw new Exception(SOSClassUtil.getMethodName() +
                          ": missing database password.");



    if ( !SOSString.isEmpty(dbname)) // falls dbname in der config-datei enthalten ist
      properties.setProperty("databasename", dbname);
    properties.setProperty("user", dbuser);
    properties.setProperty("SelectMethod", "Cursor");
    properties.setProperty("password", dbpassword);

    try {
      logger.debug9("calling " + SOSClassUtil.getMethodName());
      Driver driver = (Driver) Class.forName(this.driver).newInstance();

      connection = driver.connect(url, properties);

      if (connection == null)
        throw new Exception("can't connect to database");

      logger.debug6(".. successfully connected to " + url);
      prepare( connection) ;
    } catch (Exception e) {
      throw new Exception(SOSClassUtil.getMethodName() +
                          ": " + e.toString(),e);
    }
  }



  /**
   * prepare: Aufbereiten nach connect
   *
   * @param connection falls vorhanden ist
   * @throws java.lang.Exception falls ein Datenbankfehler vorliegt.
   */
  private void prepare( Connection connection) throws Exception {

    logger.debug6("calling " + SOSClassUtil.getMethodName());
    Statement stmt = null;
    String ISO_DATE_FORMAT = "set DATEFORMAT ymd";
    String DEFAULT_LANGUAGE = "set LANGUAGE british";
    String LOCK_TIMEOUT = "set LOCK_TIMEOUT 3000";

    try {
      if (connection == null)
        throw new Exception("can't connect to database");

      connection.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
      connection.setAutoCommit(false);

      connection.rollback();

      stmt = connection.createStatement();

      stmt.execute(ISO_DATE_FORMAT + ";" + DEFAULT_LANGUAGE + ";" + LOCK_TIMEOUT);
      logger.debug9(".. " + DEFAULT_LANGUAGE + " successfully set.");
      logger.debug9(".. " + ISO_DATE_FORMAT + " successfully set.");
      logger.debug9(".. " + LOCK_TIMEOUT + " successfully set.");
    }
    catch (Exception e) {
      throw new Exception(SOSClassUtil.getMethodName() + ": " +
                          e.toString(),e);
    }
    finally {
      try { if (stmt != null)  stmt.close(); } catch (Exception e) {}
    }
  }



  /**
   * liefert MSSQL timestamp Funktion zur�ck.
   *
   * @param dateString
   * @return oracle timestamp Funktion
   * @throws java.lang.Exception
   */
  public String toDate(String dateString) throws Exception {
    if( SOSString.isEmpty(dateString)) throw new Exception(SOSClassUtil.getMethodName() + ": dateString has no value.");
    return "{ ts '" + dateString + "'}";
  }

}
