package sos.connection;

/**
 * Title: SOSDate
 * Description:
 * Copyright: Copyright (c) 2003
 * Company: SOS GmbH
 * @author <a href="mailto:andreas.pueschel@sos-berlin.com">Andreas P�schel</a>
 */

import java.util.Properties;

import java.sql.Driver;
import java.sql.Connection;

import sos.util.SOSClassUtil;
import sos.util.SOSLogger;
import sos.util.SOSString;


/**
 * Title: 
 * <p>Description: Implementierung der SOSConnection-Klasse f�r MSSQLle</p>
 * Copyright: Copyright (c) 2003
 * Company: SOS GmbH
 * @author <a href="mailto:ghassan.beydoun@sos-berlin.com">Ghassan Beydoun</a>
 * @version $Id: SOSODBCConnection.java 937 2005-06-01 07:48:01Z gb $
 */

public class SOSODBCConnection
    extends sos.connection.SOSConnection {

  /** Umsetzung f�r %lcase, %ucase, %now */
  private static final String replacement[] = {
      "LCASE", "UCASE", "NOW()", "FOR UPDATE"};


  /**
   *
   * @param connection
   * @param logger
   * @throws java.lang.Exception
   */
  public SOSODBCConnection ( Connection connection, SOSLogger logger) throws
      Exception {
    super(connection, logger);
    super.replacement = replacement;
    prepare( connection );
  }


  /**
  *
  * @param connection
  * @throws java.lang.Exception
  */
 public SOSODBCConnection ( Connection connection) throws
     Exception {
    super(connection);
   super.replacement = replacement;
   prepare( connection) ;
 }
  
  
  /**
   * Konstruktor f�r SOSODBCConnection
   *
   * @param configFileName
   * @param logger logging Objekt
   * @throws java.lang.Exception
   * @see #SOSODBCConnection(String)
   * @see #SOSODBCConnection(String, String, String, String, String SOSLogger)
   * @see #SOSODBCConnection(String, String, String, String, String)
   */
  public SOSODBCConnection(String configFileName, SOSLogger logger) throws
      Exception {

    super(configFileName, logger);
    super.replacement = replacement;
  }


  /**
   * Konstruktor f�r SOSODBCConnection
   *
   * @param configFileName
   * @throws java.lang.Exception
   * @see #SOSODBCConnection(String, SOSLogger)
   * @see #SOSODBCConnection(String, String, String, String, String, SOSLogger)
   * @see #SOSODBCConnection(String, String, String, String, String)
   */
  public SOSODBCConnection(String configFileName) throws
      Exception {

    super(configFileName);
    super.replacement = replacement;
  }


  /**
   * Konstruktor
   *
   * @param driver
   * @param url
   * @param dbuser
   * @param dbpassword
   * @throws java.lang.Exception
   * @see #SOSMySQLConnection(String)
   * @see #SOSMySQLConnection(String, SOSLogger)
   * @see #SOSMySQLConnection(String, String, String, String, SOSLogger)
   */
  public SOSODBCConnection(String driver, String url, String dbuser, String dbpassword) throws Exception {

    super( driver, url, dbuser, dbpassword);
    super.replacement = replacement;
  }


  /**
   * Konstruktor
   *
   * @param driver
   * @param url
   * @param dbuser
   * @param dbpassword
   * @throws java.lang.Exception
   * @see #SOSMySQLConnection(String)
   * @see #SOSMySQLConnection(String, SOSLogger)
   * @see #SOSMySQLConnection(String, String, String, String, SOSLogger)
   */
  public SOSODBCConnection(String driver, String url, String dbuser, String dbpassword, SOSLogger logger) throws Exception {

    super( driver, url, dbuser, dbpassword, logger);
    super.replacement = replacement;
  }


  /**
   * Konstruktor
   *
   * @param driver
   * @param url
   * @param dbname
   * @param dbuser
   * @param dbpassword
   * @throws java.lang.Exception
   * @see #SOSODBCConnection(String)
   * @see #SOSODBCConnection(String, SOSLogger)
   * @see #SOSODBCConnection(String, String, String, String, String)
   */
  public SOSODBCConnection(String driver, String url, String dbname,
                            String dbuser, String dbpassword, SOSLogger logger) throws
      Exception {

    super(driver, url, dbuser, dbpassword, logger);
    super.replacement = replacement;
    if (dbname == null)
      throw new Exception(SOSClassUtil.getMethodName() +
                          ": missing database name.");
    this.dbname = dbname;
  }

  /**
   * Konstruktor
   *
   * @param driver
   * @param url
   * @param dbname
   * @param dbuser
   * @param dbpassword
   * @throws java.lang.Exception
   * @see #SOSODBCConnection(String)
   * @see #SOSODBCConnection(String, SOSLogger)
   * @see #SOSODBCConnection(String, String, String, String, String, SOSLogger)
   */
  public SOSODBCConnection(String driver, String url, String dbname,
                            String dbuser, String dbpassword) throws
      Exception {

    super(driver, url, dbuser, dbpassword);
    super.replacement = replacement;

    if (dbname == null)
      throw new Exception(SOSClassUtil.getMethodName() +
                          ": missing database name.");
    this.dbname = dbname;
  }


  /**
   * stellt die Datenbankverbindung her
   *
   * @throws java.lang.Exception falls ein Datenbankfehler vorliegt.
   */
  public void connect() throws Exception {

    Properties properties = new Properties();

    if ( SOSString.isEmpty(url))
      throw new Exception(SOSClassUtil.getMethodName() +
                          ": missing database url.");
    if ( SOSString.isEmpty(driver) )
      throw new Exception(SOSClassUtil.getMethodName() +
                          ": missing database driver.");
    if ( SOSString.isEmpty(dbuser))
      throw new Exception(SOSClassUtil.getMethodName() +
                          ": missing database user.");
    if ( SOSString.isEmpty(dbpassword))
      throw new Exception(SOSClassUtil.getMethodName() +
                          ": missing database password.");



/*
    String ISO_DATE_FORMAT  = "set DATEFORMAT ymd";
    String DEFAULT_LANGUAGE = "set LANGUAGE british";
*/
    if ( !SOSString.isEmpty(dbname)) // falls dbname in der config-datei enthalten ist
      properties.setProperty("databasename", dbname);
    properties.setProperty("user", dbuser);
    properties.setProperty("password", dbpassword);

    try {
      logger.debug9("calling " + SOSClassUtil.getMethodName());
      Driver driver = (Driver) Class.forName(this.driver).newInstance();

      connection = driver.connect(url, properties);

      if (connection == null)
        throw new Exception("can't connect to database");

      logger.debug6(".. successfully connected to " + url);
      prepare( connection );

/*
      stmt = connection.createStatement();
      stmt.execute(ISO_DATE_FORMAT + ";" + DEFAULT_LANGUAGE);
      logger.debug9(".. " + DEFAULT_LANGUAGE + " successfully set.");
      logger.debug9(".. " + ISO_DATE_FORMAT + " successfully set.");
*/
    } catch (Exception e) {
      throw new Exception(SOSClassUtil.getMethodName() +
                          ": " + e.toString(),e);
    }
    finally {
      // if (stmt != null) try { stmt.close(); } catch(Exception e){}
    }
  }



  /**
   * prepare: Aufbereiten nach connect
   *
   * @param connection falls vorhanden ist
   * @throws java.lang.Exception falls ein Datenbankfehler vorliegt.
   */
  private void prepare( Connection connection) throws Exception {

    try {
      logger.debug6("calling " + SOSClassUtil.getMethodName());

      connection.setAutoCommit(false);
      connection.rollback();

    }  catch (Exception e) {
      throw new Exception(SOSClassUtil.getMethodName() + ": " +
                        e.toString(),e);
    }
  }


  /**
   * liefert ODBC timestamp Funktion zur�ck.
   *
   * @param dateString
   * @return ODBC timestamp Funktion
   * @throws java.lang.Exception
   */
  public String toDate(String dateString) throws Exception {
    if( SOSString.isEmpty(dateString)) throw new Exception(SOSClassUtil.getMethodName() + ": dateString has no value.");
    return "{ ts '" + dateString + "'}";
  }

}
