package sos.scheduler.job;

import java.util.Properties;

import sos.connection.SOSConnection;
import sos.settings.SOSConnectionSettings;
import sos.settings.SOSProfileSettings;
import sos.settings.SOSSettings;
import sos.spooler.Job_impl;
import sos.util.SOSArguments;
import sos.util.SOSLogger;
import sos.util.SOSSchedulerLogger;

/**
 * Basisklasse fr Scheduler Jobs
 * 
 * @author Andreas Liebert
 */
public class JobSchedulerJob extends Job_impl {
	
    /** Protokollierung */
    private SOSSchedulerLogger sosLogger = null;
    
    /** Verbindung zur Datenbank */
    private SOSConnection sosConnection = null;
    
    /** Einstellungen aus der Datenbank */
    private SOSConnectionSettings connectionSettings = null; 
    
    /** Einstellungen des Jobs */
    private SOSSettings jobSettings = null;

    /** Job-Einstellungen */
    private Properties jobProperties = null;
    
    
	/*
	 * (non-Javadoc)
	 * 
	 * @see sos.spooler.Job_impl#spooler_init()
	 */
	public boolean spooler_init() {
	    
		try {
			boolean rc = super.spooler_init();
			if (!rc) return false;
			this.setLogger(new SOSSchedulerLogger(spooler_log));			

            try { // to initialize database connection
            	
                this.setJobSettings(new SOSProfileSettings(spooler.ini_path()));
                this.setJobProperties(this.jobSettings.getSection("spooler"));

                if (this.getJobProperties().isEmpty()) 
                    throw new Exception("no settings found in section [spooler] of configuration file: " + spooler.ini_path());

                if (this.getJobProperties().getProperty("db") == null || this.getJobProperties().getProperty("db").length() == 0) 
                    throw new Exception("no settings found for entry [db] in section [spooler] of configuration file: " + spooler.ini_path());
                
                if (this.getJobProperties().getProperty("db_class") == null || this.getJobProperties().getProperty("db_class").length() == 0) 
                    throw new Exception("no settings found for entry [db_class] in section [spooler] of configuration file: " + spooler.ini_path());
                
                
                if (this.getLogger() != null) sosLogger.debug6("connecting to database.. .");

                String dbProperty = this.getJobProperties().getProperty("db").replaceAll("jdbc:", "-url=jdbc:");
                dbProperty = dbProperty.substring(dbProperty.indexOf('-'));
                
                SOSArguments arguments = new SOSArguments(dbProperty);

	            this.setConnection( SOSConnection.createInstance(  
	                    							this.getJobProperties().getProperty("db_class"),
	                    							arguments.as_string("-class=", ""),
	                    							arguments.as_string("-url=", ""),
	                    							arguments.as_string("-user=", ""),
	                    							arguments.as_string("-password=", ""),
	                    							(SOSLogger)new SOSSchedulerLogger(this.spooler_log) )
	                    			);
	            this.getConnection().connect();
                
	            this.setConnectionSettings(new SOSConnectionSettings(this.getConnection(), "SETTINGS", this.getLogger()));
	            
	            if (this.getLogger() != null) this.getLogger().debug6("..successfully connected to " + arguments.as_string("-url=", ""));
            }
            catch (Exception e) {
                spooler_log.info("connect to database failed: " + e.getMessage());
                spooler_log.info("running without database...");
            }            
            if (spooler_job!=null) setJobProperties(getJobSettings().getSection("job " + spooler_job.name()));
            return true;
		}
		catch (Exception e) {
            if (this.getLogger() != null) try { this.getLogger().error(e.getMessage()); } catch (Exception ex) {}
            return false;
        }
	}
	
	/**
	 * @return Returns the jobSettings.
	 */
    public SOSSettings getJobSettings() {
        return this.jobSettings;
    }
    
    /**
	 * @param jobSettings
	 *            The jobSettings to set.
	 */
    public void setJobSettings(SOSSettings jobSettings) {
        this.jobSettings = jobSettings;
    }
    
    /**
	 * @return Returns the sosConnection.
	 */
    public SOSConnection getConnection() {
        return this.sosConnection;
    }
    
    /**
	 * @param sosConnection
	 *            The sosConnection to set.
	 */
    public void setConnection(SOSConnection sosConnection) {
        this.sosConnection = sosConnection;
    }
        
    /**
	 * @return Returns the sosLogger.
	 */
    public SOSSchedulerLogger getLogger() {
        return sosLogger;
    }
    
    /**
	 * @param sosLogger
	 *            The sosLogger to set.
	 */
    public void setLogger(SOSSchedulerLogger sosLogger) {
        this.sosLogger = sosLogger;
    }
    
    /**
	 * @return Returns the jobProperties.
	 */
    public Properties getJobProperties() {
        return this.jobProperties;
    }
    
    /**
	 * @param jobProperties
	 *            The jobProperties to set.
	 */
    public void setJobProperties(Properties jobProperties) {
        this.jobProperties = jobProperties;
    }

    /**
     * @return Returns the connectionSettings.
     */
    public SOSConnectionSettings getConnectionSettings() {
        return connectionSettings;
    }
    
    /**
     * @param connectionSettings The connectionSettings to set.
     */
    public void setConnectionSettings(SOSConnectionSettings connectionSettings) {
        this.connectionSettings = connectionSettings;
    }
}
