/*
 * JobSchedulerManagedDBReportJob.java
 * Created on 20.06.2005
 * 
 */
package sos.scheduler.managed;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;

import javax.xml.transform.stream.StreamResult;

import sos.connection.SOSConnection;
import sos.net.SOSMail;
import sos.spooler.Mail;
import sos.spooler.Order;
import sos.util.SOSDate;
import sos.xml.SOSXMLTransformer;

/**
 * Diese Klasse fhrt Datanbank-Statements fr Managed Orders aus und
 * verschickt einen Report im XML Format. Optional kann das xml mit Hilfe
 * eines Stylesheets in ein anderes Format gewandelt werden.<br>
 * Parameter:<br>
 * <ul>
 *   <li><strong>scheduler_order_report_stylesheet<strong>: Pfad zu einem Stylesheet, das
 * 		fr die xslt Transformation verwendet werden soll</li>
 *   <li><strong>scheduler_order_report_mailto<strong>: email-Empfnger des
 * 		 Reports.</li>
 *   <li><strong>scheduler_order_report_mailcc<strong>: email-cc-Empfnger des
 * 		 Reports.</li>
 *   <li><strong>scheduler_order_report_mailbcc<strong>: email-bcc-Empfnger des
 * 		 Reports.</li>
 *   <li><strong>scheduler_order_report_subject<strong>: email-subject des
 * 		 Reports.</li>
 *   <li><strong>scheduler_order_report_body<strong>: email-body des
 * 		 Reports.</li>
 *   <li><strong>scheduler_order_report_asbody<strong>: Bei 1 wird der Report nicht
 * 		als Attachment, sondern als body der email verschickt.</li>
 * 	 <li><strong>scheduler_order_report_filename<strong>: Der gegebene Dateiname wird 
 *      fr den Namen des Attachments verwendet und fr die permanent
 *      abgelegte Datei, falls ein Report-Pfad angegeben wird.</li>
 *   <li><strong>scheduler_order_report_path<strong>: Wird hier ein Pfad angegeben, so 
 *      wird der Report nicht nur per email verschickt, sondern auch hier abgelegt.</li>
 * </ul>
 * @author Andreas Liebert 
 */
public class JobSchedulerManagedDBReportJob extends
		JobSchedulerManagedDatabaseJob {
	
	
	private String xml;
	private String body="";
	private String mailServer;
	private String logMailFrom;
	private SOSMail mail;
	private String sql;
	
	public boolean spooler_init() {
		boolean rc =  super.spooler_init();
		try{
			Properties spoolProp = getJobSettings().getSection("spooler");
			mailServer = spoolProp.getProperty("smtp");
			logMailFrom = spoolProp.getProperty("log_mail_from");
			//mailServer = spooler_log.mail().smtp();
			//getLogger().debug9("mail().smtp():"+mailServer);
			//logMailFrom = spooler_log.mail().from();
			//getLogger().debug9("mail().from():"+logMailFrom);
		} catch (Exception e){}
		try{
			mail = new SOSMail(mailServer);
			getLogger().debug9("Setting mail sender: "+logMailFrom);
			mail.setFrom(logMailFrom);
		} catch(Exception e){
			try{getLogger().warn("Failed to initialize SOSMail class.");}catch(Exception ex){}
			rc=false;
		}
		return rc;
	}
	
	protected void executeStatements(SOSConnection conn, String command)
			throws Exception {

		File attach;
		String fileName="report_[date]_[taskid].xml";
		sql = command;
		body = "Report for statement:\n[sql]";
		boolean asbody=false;
		String subject = "Database Report [taskid]";
		
		try{
			conn.executeStatements(command);			
			conn.commit();	
		
		HashMap results = conn.get();
		if (results==null){
			getLogger().info("No results found for query");
			return;
		}
		xml = "<?xml version=\"1.0\" ?>\n";
		xml +="<report>\n";
		xml +=" <table>\n";
		xml +="  <columns>\n";
		
		Iterator keysIt = results.keySet().iterator();
		while (keysIt.hasNext()){
			xml +="<column name=\""+normalize(keysIt.next().toString())+"\"/>\n";
		}
		xml +="  </columns>\n";
		xml +="  <rows>\n";
		
		while(results!=null && !results.isEmpty()){			
			xml+="   <row>\n";
			String debug =writeFields(results);	
			xml+= debug;
			//getLogger().debug9(debug);
			xml+="   </row>\n";
			results=conn.get();
		}
		conn.closeQuery();
		conn.commit();
		xml +="  </rows>\n";
		xml +=" </table>\n";
		xml +="</report>";
		getLogger().debug3("Xml generated.");
		
		
				
		
		if (getOrderPayload()!=null && getOrderPayload().var("scheduler_order_report_asbody")!=null){
			asbody= (
					getOrderPayload().var("scheduler_order_report_asbody").equals("1") ||
					getOrderPayload().var("scheduler_order_report_asbody").equalsIgnoreCase("true"));
			debugParamter(getOrderPayload(),"scheduler_order_report_asbody");
		}
		if (getOrderPayload()!=null && getOrderPayload().var("scheduler_order_report_mailto")!=null &&
				getOrderPayload().var("scheduler_order_report_mailto").length()>0){
			String[] mails = getOrderPayload().var("scheduler_order_report_mailto").split(",");
			for (int i=0; i<mails.length; i++){
				mail.addRecipient(mails[i].trim());
			}			
			debugParamter(getOrderPayload(),"scheduler_order_report_mailto");
		}
		if (getOrderPayload()!=null && getOrderPayload().var("scheduler_order_report_mailcc")!=null &&
				getOrderPayload().var("scheduler_order_report_mailcc").length()>0){
			String[] mails = getOrderPayload().var("scheduler_order_report_mailcc").split(",");
			for (int i=0; i<mails.length; i++){
				mail.addCC(mails[i].trim());
			}
			debugParamter(getOrderPayload(),"scheduler_order_report_mailcc");
		}
		if (getOrderPayload()!=null && getOrderPayload().var("scheduler_order_report_mailbcc")!=null &&
				getOrderPayload().var("scheduler_order_report_mailbcc").length()>0){
			String[] mails = getOrderPayload().var("scheduler_order_report_mailbcc").split(",");
			for (int i=0; i<mails.length; i++){
				mail.addBCC(mails[i].trim());
			}			
			debugParamter(getOrderPayload(),"scheduler_order_report_mailbcc");
		}
		if (getOrderPayload()!=null && getOrderPayload().var("scheduler_order_report_subject")!=null &&
				getOrderPayload().var("scheduler_order_report_subject").length()>0){
			subject=(getOrderPayload().var("scheduler_order_report_subject"));
			debugParamter(getOrderPayload(),"scheduler_order_report_subject");
		}
		subject = replace(subject);
		if (getOrderPayload()!=null && getOrderPayload().var("scheduler_order_report_body")!=null &&
				getOrderPayload().var("scheduler_order_report_body").length()>0){
			body=(getOrderPayload().var("scheduler_order_report_body"));
			debugParamter(getOrderPayload(),"scheduler_order_report_body");
		}
		body=replace(body);
		if (getOrderPayload()!=null && getOrderPayload().var("scheduler_order_report_filename")!=null &&
				getOrderPayload().var("scheduler_order_report_filename").length()>0){
			fileName=getOrderPayload().var("scheduler_order_report_filename");
			debugParamter(getOrderPayload(),"scheduler_order_report_filename");			
		}		
		fileName=replace(fileName);
		getLogger().debug6("Output filename: "+fileName);
		if (getOrderPayload()!=null && getOrderPayload().var("scheduler_order_report_path")!=null && 
				getOrderPayload().var("scheduler_order_report_path").length()>0){
			String path = getOrderPayload().var("scheduler_order_report_path");
			debugParamter(getOrderPayload(),"scheduler_order_report_path");
			if (!path.endsWith("/") && !path.endsWith("\\") && path.length()>0) path = path +"/";
			path=replace(path);
			File reportPath = new File(path);
			if(!reportPath.exists()) reportPath.mkdirs();
			attach = new File(path+fileName);
		}else{
			String path = System.getProperty("java.io.tmpdir");
			if (!path.endsWith("/") && !path.endsWith("\\") && path.length()>0) path = path +"/";
			attach = new File(path+fileName);
			attach.deleteOnExit();
		}		
		
		if(getOrderPayload()!=null && getOrderPayload().var("scheduler_order_report_stylesheet")!=null &&
				getOrderPayload().var("scheduler_order_report_stylesheet").length()>0){
			debugParamter(getOrderPayload(),"scheduler_order_report_stylesheet");
			File stylesheet = new File(getOrderPayload().var("scheduler_order_report_stylesheet"));
			if (!stylesheet.canRead()){
				throw new Exception("Could not read stylesheet: "+stylesheet.getAbsolutePath());				
			}else{
				getLogger().debug3("Doing xslt transformation...");
				//ByteArrayOutputStream bos = new ByteArrayOutputStream();
				//StreamResult output = new StreamResult(bos);
				try{
					getLogger().debug6("attach:"+attach.getAbsolutePath());
					SOSXMLTransformer.transform(xml, stylesheet, attach);
				} catch(Exception e){
					getLogger().error("Error occured during transformation: "+e);					
				}
				getLogger().debug3("Xslt transformation done.");
				if(asbody){
					BufferedReader reader = new BufferedReader(new FileReader(attach));
					StringBuffer buffer = new StringBuffer();
					String line = reader.readLine();
					while (line != null) {
						buffer.append(line);
						buffer.append('\n');
						line = reader.readLine();
					}
					reader.close();
					body = buffer.toString();
					mail.setContentType("text/html");
				}
			}
		} else{
			FileOutputStream fos = new FileOutputStream(attach);
			fos.write(xml.getBytes());
			fos.flush();
			fos.close();
			if (asbody) body=xml;
		}
		if(body.trim().startsWith("<html")) mail.setContentType("text/html");
		mail.setBody(body);	
		mail.setSubject(subject);
		if (!asbody) mail.addAttachment(attach.getAbsolutePath());
		mail.send();
		} catch(Exception e){
			Order order = spooler_task.order();
			getLogger().warn("An error occured creating database report"+ ((order != null) ? "  [Job Chain: " + order.job_chain().name() + ", Order:"+ order.id()+"]" : "")+": "+e);
			spooler_task.end();
		} finally{
			mail.clearAttachments();
			mail.clearRecipients();
		}
		
	}
	
	
		
	private String normalize(String text){
		return text;
	}
	
	private String writeFields(HashMap results){
		Iterator keysIt = results.keySet().iterator();
		String rc = "";
		while(keysIt.hasNext()){
			String key = keysIt.next().toString();
			String value = results.get(key).toString();
			if(value==null || value.length()==0)
				rc += "    <" +normalize(key)+"/>";
			else
				rc += "    <" +normalize(key)+"><![CDATA["+value+"]]></"+normalize(key)+">\n";
		}
		return rc;
	}
	
	private String replace(String source){
		String target;
		try{
			target = source.replaceAll("\\[date\\]",SOSDate.getCurrentDateAsString());
			target = target.replaceAll("\\[datetime\\]",SOSDate.getCurrentTimeAsString());
			if(isOrderJob())
			target = target.replaceAll("\\[orderid\\]", spooler_task.order().id());
			target = target.replaceAll("\\[jobname\\]", spooler_job.name());
			target = target.replaceAll("\\[taskid\\]", Integer.toString(spooler_task.id()));
			target = target.replaceAll("\\[sql\\]",sql);
			target = target.replaceAll("\\[xml\\]",xml);
			
		} catch (Exception e){
			try{
				getLogger().warn("An error occured replacing fields in String \""+source+"\"");
			} catch(Exception ex){}
			return source;
		}
		return target;
	}
}
