package sos.scheduler.managed;

import java.io.ByteArrayInputStream;
import java.util.HashMap;


import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import sos.connection.SOSConnection;
import sos.spooler.Job_chain;
import sos.spooler.Job_impl;
import sos.spooler.Order;
import sos.spooler.Variable_set;
import sos.util.SOSLogger;
import sos.spooler.Log;
import sos.util.SOSSchedulerLogger;

/**
 * Statische Hilfklasse fr Managed Jobs
 * @author Andreas Pschel <andreas.pueschel@sos-berlin.com>
 * @since 2005-03-05
 * @version 1.0
 */
public class JobSchedulerManagedObject {

    /** Tabelle der persistenten Auftrge */
    private static String tableManagedOrders = "SCHEDULER_MANAGED_ORDERS";
    
    /** Tabelle der persistenten Auftragsparameter */
    private static String tableManagedOrderParameters = "SCHEDULER_MANAGED_ORDER_PARAMETERS";
    
    /** Tabelle der Datenbankverbindungen */
    private static String tableManagedConnections = "SCHEDULER_MANAGED_CONNECTIONS";
    
    /** Tabelle der Job-Typen */
    private static String tableManagedJobTypes = "SCHEDULER_MANAGED_JOB_TYPES";
    
    /** Tabelle der Jobs */
    private static String tableManagedJobs = "SCHEDULER_MANAGED_JOBS";
    
    /** Tabelle der Benutzer-Auftrge */
    private static String tableManagedUserJobs = "SCHEDULER_MANAGED_USER_JOBS";
    
    /** Tabelle der Temporren Benutzer fr die User Jobs */
    private static String tableManagedTempUsers = "SCHEDULER_MANAGED_TEMP_USERS";
    
    /** Tabelle der Workflow-Modelle */
    private static String tableManagedModels = "SCHEDULER_MANAGED_MODELS";
    

    /** Tabelle der Workflow-Pakete */
    private static String tableWorkflowPackages = "WORKFLOW_PACKAGES";
    
    /** Tabelle der Workflow-Historie */
    private static String tableWorkflowHistory = "WORKFLOW_HISTORYS";
    
    
    
    
    /**
     * @return Returns the tableManagedOrderParameters.
     */
    public static String getTableManagedOrderParameters() {
        return tableManagedOrderParameters;
    }
    
    /**
     * @param tableManagedOrderParameters The tableManagedOrderParameters to set.
     */
    public static void setTableManagedOrderParameters(String tableManagedOrderParameters) {
        JobSchedulerManagedObject.tableManagedOrderParameters = tableManagedOrderParameters;
    }
    
    /**
     * @return Returns the tableManagedOrders.
     */
    public static String getTableManagedOrders() {
        return tableManagedOrders;
    }
    
    /**
     * @param tableManagedOrders The tableManagedOrders to set.
     */
    public static void setTableManagedOrders(String tableManagedOrders) {
        JobSchedulerManagedObject.tableManagedOrders = tableManagedOrders;
    }
    
    /**
     * @return Returns the tableWorkflowHistory.
     */
    public static String getTableWorkflowHistory() {
        return tableWorkflowHistory;
    }
    
    /**
     * @param tableWorkflowHistory The tableWorkflowHistory to set.
     */
    public static void setTableWorkflowHistory(String tableWorkflowHistory) {
        JobSchedulerManagedObject.tableWorkflowHistory = tableWorkflowHistory;
    }
    
    /**
     * @return Returns the tableManagedModels.
     */
    public static String getTableManagedModels() {
        return tableManagedModels;
    }
    
    /**
     * @param tableWorkflowModels The tableManagedModels to set.
     */
    public static void setTableManagedModels(String tableManagedModels) {
        JobSchedulerManagedObject.tableManagedModels = tableManagedModels;
    }
    
    /**
     * @return Returns the tableWorkflowPackages.
     */
    public static String getTableWorkflowPackages() {
        return tableWorkflowPackages;
    }
    
    /**
     * @param tableWorkflowPackages The tableWorkflowPackages to set.
     */
    public static void setTableWorkflowPackages(String tableWorkflowPackages) {
        JobSchedulerManagedObject.tableWorkflowPackages = tableWorkflowPackages;
    }
    
    /**
     * @return Returns the tableManagedConnections.
     */
    public static String getTableManagedConnections() {
        return tableManagedConnections;
    }
    
    /**
     * @param tableManagedConnections The tableManagedConnections to set.
     */
    public static void setTableManagedConnections(String tableManagedConnections) {
        JobSchedulerManagedObject.tableManagedConnections = tableManagedConnections;
    }
    
    /**
     * @return Returns the tableManagedJobs.
     */
    public static String getTableManagedJobs() {
        return tableManagedJobs;
    }
    
    /**
     * @param tableManagedJobs The tableManagedJobs to set.
     */
    public static void setTableManagedJobs(String tableManagedJobs) {
        JobSchedulerManagedObject.tableManagedJobs = tableManagedJobs;
    }
    
    /**
     * @return Returns the tableManagedJobTypes.
     */
    public static String getTableManagedJobTypes() {
        return tableManagedJobTypes;
    }
    
    /**
     * @param tableManagedJobTypes The tableManagedJobTypes to set.
     */
    public static void setTableManagedJobTypes(String tableManagedJobTypes) {
        JobSchedulerManagedObject.tableManagedJobTypes = tableManagedJobTypes;
    }
    
    /**
     * @return Returns the tableManagedUserJobs.
     */
    public static String getTableManagedUserJobs() {
        return tableManagedUserJobs;
    }
    /**
     * @param tableManagedUserJobs The tableManagedUserJobs to set.
     */
    public static void setTableManagedUserJobs(String tableManagedUserJobs) {
        JobSchedulerManagedObject.tableManagedUserJobs = tableManagedUserJobs;
    }

    /**
     * Gibt ein Connection Object zurck, welches die Order-spezifische Verbindung darstellt
     * @param conn Connection, die benutzt wird, um die Daten fr die neue Connection zu finden
     * @param job job Object, von dessen Order die Connection besorgt werden soll (meistens "this")
     * @return Connection, die zum Order gehrt. Diese ist noch nicht initialisiert, es mssen also
     * noch connect() und spter disconnect() aufgerufen werden.
     * @throws Exception
     */
    public static SOSConnection getOrderConnection(SOSConnection connection, Job_impl job) throws Exception {
        
    	SOSConnection localConnection = null;
        Order order = null;
        
        boolean rc = false;
        HashMap result = null;
        String spoolerId = job.spooler.id().toLowerCase();
        String connectionName = "";
        
        if(job.spooler_task.job().order_queue()!=null){
        	order = job.spooler_task.order(); 
        	String jobChain = order.job_chain().name();
        	String jobChainModel = "";
        	String orderId = getOrderIdInTable(spoolerId, jobChain, order);
        	try{
        		sos.spooler.Variable_set orderPayload = (sos.spooler.Variable_set) order.payload();
    	
        		job.spooler_log.debug3("Oder Payload: "+orderPayload.xml());
   	   			connectionName=orderPayload.var("database_connection");
        	} catch (Exception e){}
   	   	
   	   		job.spooler_log.debug3("Connection from payload:"+connectionName);
   	    
   	   		if (connectionName == null || connectionName.length() == 0) {
   	   			connectionName = connection.getSingleValue("SELECT \"CONNECTION\" FROM " + getTableManagedOrders() +
   	    			" WHERE \"SPOOLER_ID\"='" + spoolerId + "' AND \"JOB_CHAIN\"='" + jobChain + "'" +
   					" AND \"ORDER_ID\"='" + orderId + "'");
   	   		}
    	
   	   		if (connectionName == null || connectionName.length() == 0) {
   	   			connectionName = connection.getSingleValue("SELECT \"CONNECTION\" FROM " + getTableManagedOrders() +
        			" WHERE \"SPOOLER_ID\" IS NULL AND \"JOB_CHAIN\"='" + jobChain + "' AND " +
        			"\"ORDER_ID\"='" + orderId + "'");
   	   		}
    	
   	   		if (connectionName == null || connectionName.length() == 0) {
   	   			jobChainModel = connection.getSingleValue("SELECT \"ID\" FROM " + getTableManagedModels() +
    	            " WHERE \"NAME\"='" + jobChain + "'");
    	    
        		if (jobChainModel == null || jobChainModel.length() == 0)
          		throw (new Exception("no model found for job chain [" + jobChain + "]of managed order: " + orderId));
   	   		}
   	   		
        }
        	
    	if (connectionName == null || connectionName.length() == 0) {
       	    connectionName = connection.getSingleValue("SELECT \"CONNECTION\" FROM " + getTableManagedJobs() +
        			" WHERE \"SPOOLER_ID\"='" + spoolerId + "' AND \"JOB_NAME\"='" + job.spooler_job.name() + "'");
        }
    	
    	if (connectionName == null || connectionName.length() == 0) {
       		connectionName = connection.getSingleValue("SELECT \"CONNECTION\" FROM " + getTableManagedJobs() +
           			" WHERE \"SPOOLER_ID\" IS NULL AND " +
           			"\"JOB_NAME\"='" + job.spooler_job.name() + "'");
    	}
    	
    	if (connectionName == null || connectionName.length() == 0)
      		throw (new Exception("no database connection identifier found for managed order"));
    	
    	result = connection.getSingle("SELECT \"DRIVER\", \"CLASS\", \"URL\", \"USERNAME\", \"PASSWORD\", \"CONNECTION\" FROM "
		        + JobSchedulerManagedObject.getTableManagedConnections() 
		        + " WHERE \"CONNECTION\"='" + connectionName + "'");
    	if (result.isEmpty()) throw (new Exception("no connection settings found for managed connection: " 
		        + connectionName)); 
		
		
	    try {
            job.spooler_log.debug6("..creating local connection object");
            
            localConnection = SOSConnection.createInstance(  
                    result.get("class").toString(),
                    result.get("driver").toString(),
                    result.get("url").toString(),
                    result.get("username").toString(),
                    result.get("password").toString(),
					(SOSLogger)new SOSSchedulerLogger(job.spooler_log) );
            
	    }
	    catch (Exception e) {
	        throw new Exception("error occurred establishing database connection: " + e.getMessage());
	    }
	    
    	return localConnection;
    }

    
    /**
     * Liefert den Job-spezifischen command zurck
     * @param connection Connection, die benutzt wird, um die Daten fr den Command zu finden
     * @param job job Object, von dessen Order der Command besorgt werden soll (meistens "this")
     * 
     * @throws Exception
     */
    public static String getJobCommand(SOSConnection connection, Job_impl job) throws Exception{
    	
		String command = "";
		try{
		sos.spooler.Variable_set jobPayload = (sos.spooler.Variable_set) job.spooler_task.params();
		command=jobPayload.var("command");
		} catch (Exception e){}
		job.spooler_log.debug3("job command: " + command);
    	job.spooler_log.debug3("command.lentgh: "+ command.length());
		if(command!=null && command.length()>0) {
			String realCommand = new String(fromHexString(command),"US-ASCII");
			return realCommand;
		}
		

		if (job.spooler.id() != null && job.spooler.id().length() > 0) {
    	    command = connection.getClob("SELECT \"COMMAND\" FROM " + getTableManagedJobs() +
    			" WHERE \"SPOOLER_ID\"='" + job.spooler.id().toLowerCase() + "'" +
    			" AND \"JOB_NAME\"='" + job.spooler_job.name() + "'");
		} 
		
		if (command == null || command.length() == 0) {
    		command = connection.getClob("SELECT \"COMMAND\" FROM "+getTableManagedJobs() +
        			" WHERE \"SPOOLER_ID\" IS NULL" +
        			" AND \"JOB_NAME\"='" + job.spooler_job.name() + "'");
    	}

    	job.spooler_log.debug3("job command: " + command);

    	return command;
    }

    
    /**
     * Liefert den Order-spezifischen command zurck
     * @param conn Connection, die benutzt wird, um die Daten fr den Command zu finden
     * @param job job Object, von dessen Order der Command besorgt werden soll (meistens "this")
     * 
     * @throws Exception
     */
    public static String getOrderCommand(SOSConnection connection, Job_impl job) throws Exception{
    	    	
    	job.spooler_log.debug9("entered getOrderCommand()...");
    	Order order = job.spooler_task.order();
    	job.spooler_log.debug9("order!=null: "+(order!=null));
    	String spoolerID = job.spooler.id().toLowerCase();
    	job.spooler_log.debug9("spoolerID: "+spoolerID);
    	Job_chain chain = order.job_chain();
    	job.spooler_log.debug9("chain!=null: "+(chain!=null));
    	String jobChainName = chain.name();
    	job.spooler_log.debug9("jobChainName: "+jobChainName);
    	String orderID = getOrderIdInTable(spoolerID, jobChainName, order);
    	
		String command = "";
		try{
			job.spooler_log.debug9("trying to get Command from order.payload()...");
			sos.spooler.Variable_set orderPayload = (sos.spooler.Variable_set) order.payload();
    	
			command=orderPayload.var("command");			
		} catch (Exception e) {}
		
		if(command!=null && command.length()>0) {
			String realCommand = new String(fromHexString(command),"US-ASCII");
			return realCommand;
		}
		
		job.spooler_log.debug9("trying to get Command from table "+ getTableManagedOrders()+" ...");
    	if (!order.id().startsWith("-")) {
    	    command = connection.getClob("SELECT \"COMMAND\" FROM " + getTableManagedOrders() +
    			" WHERE \"SPOOLER_ID\"='" + job.spooler.id().toLowerCase() + "' AND \"JOB_CHAIN\"='" + order.job_chain().name() + "'" +
    			" AND \"ORDER_ID\"='" + orderID + "'");
    	} else {
    		command = connection.getClob("SELECT \"COMMAND\" FROM " + getTableManagedOrders() +
        		" WHERE \"SPOOLER_ID\" IS NULL AND \"JOB_CHAIN\"='" + order.job_chain().name() + "'" +
        		" AND \"ORDER_ID\"='" + orderID + "'");
    	}
    	
    	job.spooler_log.debug3("order command: " + command);
    	job.spooler_log.debug3("command.lentgh: "+ command.length());

    	return command;
    }

    
    private static String substringAfter(String str, String separator) {
        if ((str.length()==0)) {
            return str;
        }
        if (separator == null) {
            return "";
        }
        int pos = str.indexOf(separator);
        if (pos == -1) {
            return "";
        }
        return str.substring(pos + separator.length());
    }
    
    /**
     * Wenn der Order aus der Tabelle SCHEDULER_MANAGED_ORDERS erzeugt wird, wird
     * ihm eine neue ID vergeben. Diese Funktion gibt den ursprnglichen
     * Ordernamen in der Tabelle zurck.
     * @param order
     */
    public static String getOrderIdInTable(String spoolerID, String jobChain, Order order) {
        
    	if(order.id().startsWith("-")) spoolerID="";
    	String prefix = spoolerID+"-"+jobChain+"-";
    	return substringAfter(order.id(), prefix);
    }
    
    
    public static String replaceVariablesInCommand(String command, Variable_set vars) throws Exception{
    	return replaceVariablesInCommand(command, vars, null);
    }
    
    /**
     * Ersetzt Variablen der Form ${name} im command mit entsprechenden Werten im
     * Variable_set
     * @param command Ursprnglicher command mit Platzhaltern
     * @param vars jobParameter
     * @return neuer command mit ersetzten Variablen
     */
    public static String replaceVariablesInCommand(String command, Variable_set vars, Log log) throws Exception{
    	DocumentBuilderFactory docFactory 	= DocumentBuilderFactory.newInstance();                
        DocumentBuilder docBuilder 			= docFactory.newDocumentBuilder();                
    	Document payloadDocument  			= docBuilder.parse(new ByteArrayInputStream(vars.xml().getBytes()));
    	
    	
    	Node node = payloadDocument.getFirstChild();
        while( node != null  &&  node.getNodeType() != Node.ELEMENT_NODE ) node = node.getNextSibling();

        if ( node == null ) {
            throw new Exception("payload contains no xml content");
        }

        Element payloadElement = (Element)node;
//        if (!payloadElement.getNodeName().equals( "variables" ) )  
//            throw new Exception( "element <variables> is missing" );
             
        node = payloadElement.getFirstChild();
        while( node != null  &&  node.getNodeType() != Node.ELEMENT_NODE )  node = node.getNextSibling();
 
        boolean ascending = false;
        String parameterName = "";
        String parameterValue = "";
        
        while(node != null) {
            NamedNodeMap attributes = node.getAttributes();

            if ( node.getNodeName().equals( "variable" ) ) {
                try {
                    if (attributes.getNamedItem("name") == null) {
                        throw new Exception ("attribute <variable name=\"...\"> is missing");
                    }
                    parameterName = attributes.getNamedItem("name").getNodeValue();
                    //if (log!=null) log.debug9("Name: "+parameterName);
                } catch (Exception e) {
                    throw new Exception( "illegal value for attribute [name]: " + attributes.getNamedItem("name").getNodeValue());
                }

                try {
                    if (attributes.getNamedItem("value") == null) {
                        throw new Exception ("attribute <variable value=\"...\"> is missing");
                    }
                    //parameterValue = attributes.getNamedItem("value").getNodeValue();
                    parameterValue = vars.var(parameterName).replaceAll("\\\\","\\\\\\\\");
                    //if (log!=null) log.debug9("Value: "+parameterValue);
                } catch (Exception e) {
                    throw new Exception( "illegal value for attribute: " + attributes.getNamedItem("value").getNodeValue());
                }
                
		        command = command.replaceAll("\\$\\{" + parameterName + "\\}", parameterValue.replaceAll("\\[quot\\]", "'"));
		        //command = command.replaceAll("\\$\\{" + parameterName + "\\}", parameterValue);
		        //if (log!=null) log.debug9("command: "+command);
            } else if ( node.getNodeName().equals( "sos.spooler.variable_set" ) ) {
                break;
            } 
            
            if (node.hasChildNodes() && !ascending) {
                node = node.getFirstChild();
            } else  if (node.getNextSibling() != null) {
                node = node.getNextSibling();
                ascending = false;
            } else if (node.getParentNode() != null) {
                node = node.getParentNode();
                ascending = true;
            } else {
                break;
            }
        }
    	return command;
    }
    
    /**
    * Convert a hex string to a byte array.
    * Permits upper or lower case hex.
    *
    * @param s String must have even number of characters.
    * and be formed only of digits 0-9 A-F or
    * a-f. No spaces, minus or plus signs.
    * @return corresponding byte array.
    */
    public static byte[] fromHexString ( String s )
       {
       int stringLength = s.length();
       if ( (stringLength & 0x1) != 0 )
          {
          throw new IllegalArgumentException ( "fromHexString requires an even number of hex characters" );
          }byte[] b = new byte[stringLength / 2];

       for ( int i=0,j=0; i<stringLength; i+=2,j++ )
          {
          int high = charToNibble( s.charAt ( i ) );
          int low = charToNibble( s.charAt ( i+1 ) );
          b[j] = (byte)( ( high << 4 ) | low );
          }
       return b;
       }

    /**
    * convert a single char to corresponding nibble.
    *
    * @param c char to convert. must be 0-9 a-f A-F, no
    * spaces, plus or minus signs.
    *
    * @return corresponding integer
    */
    private static int charToNibble ( char c )
       {
       if ( '0' <= c && c <= '9' )
          {
          return c - '0';
          }
       else if ( 'a' <= c && c <= 'f' )
          {
          return c - 'a' + 0xa;
          }
       else if ( 'A' <= c && c <= 'F' )
          {
          return c - 'A' + 0xa;
          }
       else
          {
          throw new IllegalArgumentException ( "Invalid hex character: " + c );
          }
       }
    
    public static String toHexString(byte[] b)
    {   
        StringBuffer buf = new StringBuffer();
        for (int i=0;i<b.length;i++)
        {
            int bi=0xff&b[i];
            int c='0'+(bi/16)%16;
            if (c>'9')
                c= 'A'+(c-'0'-10);
            buf.append((char)c);
            c='0'+bi%16;
            if (c>'9')
                c= 'a'+(c-'0'-10);
            buf.append((char)c);
        }
        return buf.toString();
    }
    
	/**
	 * @return Returns the tableManagedTempUsers.
	 */
	public static String getTableManagedTempUsers() {
		return tableManagedTempUsers;
	}
	/**
	 * @param tableManagedTempUsers The tableManagedTempUsers to set.
	 */
	public static void setTableManagedTempUsers(String tableManagedTempUsers) {
		JobSchedulerManagedObject.tableManagedTempUsers = tableManagedTempUsers;
	}
}
