package sos.settings;

import java.util.*;
import sos.util.*;
import sos.xmlparser.*;

/**
 * <p>Title: SOSXMLSettings.java </p>
 * <p>Description: Diese Klasse liest eine XML-Konfigurationsdatei aus.</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: SOS GmbH</p>
 * @author Mrvet ksz
 * @resource sos.util.jar; sos.xmlparser.DOMParserXML.class
 * @author Mrvet ksz
 */
public class SOSXMLSettings extends SOSSettings {   
    
    /** Liste von HashTable, die alle Eintrge beiinhaltet */
    private ArrayList listOfXMLPath = null;
        
    /** sos.tool.DOMParserXML Objekt */
    private DOMParserXML parser = null;
    
    /** sos.util.SOSString Objekt*/
    private SOSString sosString = null;
    
    /** Name der Anwendung */
    private String application = "";
    
    /** Liste von HashTabelle. Die HashTabelle beinhaltet die Tags mit entsprechenden Tagwerten*/
    
	/**
	 * Konstruktor
	 *
	 * @param source Name der Datenquelle
	 * @param logger Das Logger-Objekt
	 *
	 * @throws java.lang.Exception
	 */
	 public SOSXMLSettings ( String source,
										  SOSLogger logger)
		  throws Exception  {
		  super( source, logger );
		  this.load();
	 }


	 /**
	  * Konstruktor
	  *
	  * @param source Name der Datenquelle
	  * @param section Name der Sektion
	  * @param logger Das Logger-Objekt
	  *
	  * @throws java.lang.Exception
	  */
	  public SOSXMLSettings ( String source,
											String section,
											SOSLogger logger
										  )
			throws Exception  {      
			super( source, section, logger );
			this.load();
	  }

	  /**
	   * Ldt die XML-Datei in den Speicher
	   */
	  private void load() throws Exception {
	      //Hilfsvariable
	      ArrayList listOfAllTags = null;
	      HashMap currEntry = null;
	      try {
	          listOfXMLPath = new ArrayList();
	          sosString = new SOSString();
	          parser = new DOMParserXML((SOSStandardLogger)logger);
	          parser.setDepth(0); //0=nur Tagname; 1=nur der Vaterknoten;
	          // 99=gesamte Pfad
	          parser.setRemoveParents(true);
	          //parser.setTableName("tableXYZ");
	          parser.parse(source);
	          //ZUM TESTEN: Liste aller insertstatment; vorher defaults hinzufgen
	          //listofStatement = parser.getInsertStatement();
	          //for (int i = 0; i < listofStatement.size(); i++) {
	          //    System.out.println(i + ": " + listofStatement.get(i));
	          //}
	          
	          //Liste alle Tags und Tagswerte in einer Hashtabelle ausgeben
	          listOfAllTags = parser.getListOfXMLPath();
	          for (int i = 0; i < listOfAllTags.size(); i++ ) {
	              currEntry = (HashMap)listOfAllTags.get(i);
	              if (currEntry.containsKey("application") && currEntry.containsKey("section") && currEntry.containsKey("name")) {
	                  listOfXMLPath.add(currEntry);
	              }
	          }
	      } catch (Exception e) {
	          throw new Exception ("\n -> ..error in " + SOSClassUtil.getMethodName() + " " + e);
	      }
	  }

  /**
   * Liefert alle Eintrge einer Sektion zurck.
   *
   * @return Properties Objekt, das alle Eintrge der Sektion darstellt.
   * @exception Exception
   * @see #getSection( String )
   */
  public Properties getSection(String application, String section)
     throws Exception {           
      try {
          this.application = application;
          this.section = section;
          
          Properties retVal = new Properties();
          HashMap h = null;
          for (int i = 0; i < listOfXMLPath.size(); i++) {
              h = (HashMap) (listOfXMLPath.get(i));
              if (sosString.parseToString(h, "application").equalsIgnoreCase(application)
                  && sosString.parseToString(h, "section").equalsIgnoreCase(section)) {
                  retVal.put(h.get("name"), h.get("value"));
              }                                
          }
          return retVal; 
      } catch (Exception e) {
          throw new Exception ("\n -> ..error in " + SOSClassUtil.getMethodName() + " " + e);
      }
  }
  
   /** 
    * Auslesen der Sektion
    * @see sos.settings.SOSSettings#getSection()
    */
   public Properties getSection() throws Exception {
       try {
           if (SOSString.isEmpty(application)) {
               throw (new Exception(SOSClassUtil.getMethodName() + ": application has no value!"));
           }
           if (SOSString.isEmpty(section)) {
               throw (new Exception(SOSClassUtil.getMethodName() + ": section has no value!"));
           }   
               
           return getSection(application, section);
       } catch (Exception e) {
           throw new Exception ("\n -> ..error in " + SOSClassUtil.getMethodName() + " " + e);
       }
   	}

  /**
   * Liefert alle Eintrge der eingegebenen Sektion zurck.
   *
   * @return Properties Objekt, das alle Eintrge der Sektion darstellt.
   * @throws java.lang.Exception
   * @see #getSection( String )
   */
   public Properties getSection(String section) throws Exception {
       try {
           if (SOSString.isEmpty(application)) {
               throw (new Exception(SOSClassUtil.getMethodName() + ": application has no value!"));
           }                      
           return getSection(this.application, section);           
       } catch (Exception e) {
           throw new Exception ("\n -> ..error in " + SOSClassUtil.getMethodName() + " " + e);
       }      
   }

   /**
    * Liefert alle Sektionen einer Anwendung zurck
    *
    * @return ArrayList die alle Sektionen beinhaltet
    * @throws java.lang.Exception
    */
   public ArrayList getSections( String application ) throws Exception {
       ArrayList retVal = new ArrayList();
       try {
           this.application = application;           
                      
           HashMap h = null;
           for (int i = 0; i < listOfXMLPath.size(); i++) {
               h = (HashMap) (listOfXMLPath.get(i));
               if (sosString.parseToString(h,"application").equalsIgnoreCase(application)) {
                   if (!retVal.contains(sosString.parseToString(h, "section"))) {
                       retVal.add(h.get("section"));
                   }                                
               }
           }
           return retVal; 
       } catch (Exception e) {
           throw new Exception ("\n -> ..error in " + SOSClassUtil.getMethodName() + " " + e);
       }
   }
   
   /**
    * Liefert alle Sektionen einer Anwendung zurck
    *
    * @return ArrayList die alle Sektionen beinhaltet
    * @throws java.lang.Exception
    */
   public ArrayList getSections() throws Exception {
       try {
       if (SOSString.isEmpty(application)) {
           throw (new Exception(SOSClassUtil.getMethodName() + ": application has no value!"));
       }                      
       return getSections( application );
       } catch (Exception e) {
           throw new Exception ("\n -> ..error in " + SOSClassUtil.getMethodName() + " " + e);
       }
   }

    /**
   * Liefert den wert eines Eintrages zurck.
   *
   * @param entry Name des Eintrages
   * @return String der Wert eines Eintrages
   * @throws java.lang.Exception   
   */
   public String getSectionEntry(String entry) throws Exception {      
       try {
       if (SOSString.isEmpty(application)) {
           throw (new Exception(SOSClassUtil.getMethodName() + ": application has no value!"));
       }
       if (SOSString.isEmpty(section)) {
           throw (new Exception(SOSClassUtil.getMethodName() + ": section has no value!"));
       }   
       HashMap h = null;
       for (int i = 0; i < listOfXMLPath.size(); i++) {
           h = (HashMap) (listOfXMLPath.get(i));
           if (sosString.parseToString(h, "application").equalsIgnoreCase(application)
               && sosString.parseToString(h, "section").equalsIgnoreCase(section) &&
               sosString.parseToString(h, "name").equalsIgnoreCase(entry)) {
               return sosString.parseToString(h, "value");
           }                                
       }
       return "";
       } catch (Exception e) {
           throw new Exception ("\n -> ..error in " + SOSClassUtil.getMethodName() + " " + e);
       }
   }

   /* (non-Javadoc)
    * @see sos.settings.SOSSettings#setKeysToLowerCase()
    */
   public void setKeysToLowerCase() throws Exception {   }

   /* (non-Javadoc)
    * @see sos.settings.SOSSettings#setKeysToUpperCase()
    */
   public void setKeysToUpperCase() throws Exception {   }

 
    /**
     * Diese Methode liefert den Wert der Property <i>listOfXMLPath</i>. 
     * Der Typ der Property <i>listOfXMLPath</i> ist ArrayList.
     * Die ArrayList beiihaltet eine Liste von HashTable, die folgende
     * Keys hat; application, section , name und key
     * 
     * 
     * @return ArrayList.
     */
    public ArrayList getListOfXMLFields() {
        return listOfXMLPath;
    }
    
    /**
     * Setzt den Schalter fr die Bercksichtigung von Gro/Kleinschreibung
     * @param ignoreCase
     */
    public void setIgnoreCase( boolean ignoreCase ) {

      this.ignoreCase = ignoreCase;
    }
    
    /**
     * liefert den Schalter fr die Bercksichtigung von Gro/Kleinschreibung
     * @param ignoreCase
     */
    public boolean getIgnoreCase() {
      return this.ignoreCase;
    }
    
}