package sos.settings.converter;

import java.util.*;

import sos.util.*;
import sos.settings.*;
import java.io.*;


/**
 * <p>Title: sos.settings.converter.SOSSettingsProfile2XMLConverter.java </p>
 * <p>Description: Konvertiert eine vorhandene Ini-Konfigurationsdatei in XML-Konfigurationsdatei.
 * Diese XML-Konfigurationsdatei kann ber die Web-Oberflche in die Datenbank importiert werden.</p>
 * <p>Copyright: Copyright (c) 08.07.2004</p>
 * <p>Company: SOS GmbH</p>
 * @author Mrvet ksz
 * @resource 
 */
public class SOSSettingsProfile2ConnectionConverter extends SOSSettingsConverter {
    
    /**
     * Attribut: SOSStandardLogger Objekt.
     * @see sos.util.SOSStandardLogger 
     * */
    private SOSStandardLogger sosLogger = null;
    
    /** Attribut: Name der Konfigurationsdatei*/
    private String configFile = "";
        
    /** Attribut: Alle Eintrge der Sektion*/
    private ArrayList sections = null;
    
    /** Attribut: BufferWriter */
    private BufferedWriter output = null;
    
    /** Attribut: encoding; default ist iso-8859-1 */
    private String encoding = "iso-8859-1";
    
    /** Attribut: namespace*/
    private String namespace = "";
    
    /** Attribut: Name der Application */
    private String application = "application";
    
    /** Attribut: SOSString Objekt */
    private SOSString sosString = null;
    
    /** Attribut: SOSProfileSettings Object*/
    private SOSProfileSettings settings = null;
    
    /** Attribut: Name der Tabelle*/
    private String tablename = "SETTINGS"; 
    
    /**
     * @throws Exception
     */
    public SOSSettingsProfile2ConnectionConverter() throws Exception {
        super();        
    }

    /**
     * @param sosLogger_
     * @throws Exception
     */
    public SOSSettingsProfile2ConnectionConverter(SOSStandardLogger sosLogger_)
            throws Exception {
        super(sosLogger_);
    }
    
    /**
     * Konstruktor.
     * 
     * @param SOSStandardLogger
     * @see sos.util.SOSStandardLogger
     * @throws Exception
     */
    public SOSSettingsProfile2ConnectionConverter(String configFile_, SOSStandardLogger sosLogger_)
    		throws Exception {
        super(configFile_, sosLogger_);
        
    }
    
    /**
     * Ldt die INI-Konfigurationsdatei. <br>
     * Diese Methode liefert true, wenn die Konfigurationsdatei erfolgreich 
     * geladen wurde. 
     * 
     * @return boolean
     * @throws Exception
     */
    public boolean loadSettings() throws Exception {
        boolean retVal  =true;        
        try {            
	        settings = new SOSProfileSettings(configFile, sosLogger);  
	        sections = settings.getSections();
            return retVal;
        } catch (Exception e) {
            throw new Exception("\n ->..error in " + SOSClassUtil.getMethodName() + " " + e);
        }
    }
    
    /**
     * Schreibt die INI-Konfigurationsdatei in eine XML-Konfigutaionsdatei fr die 
     * SOSConnectionSettings.<br> 
     * z.B. von ini zu xml Datei.<br><br>
     * 
     * Die XML-Datei sieht z.B. folgendermaen aus:<br>
     * <?xml version="1.0" encoding="UTF-8"?><br>
     * <sos_export xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" 
     * 		xsi:noNamespaceSchemaLocation="C:\eclipse\workspace\sos.tool\sos\settings\Untitled1.xsd" application="String" created="String"><br>
     * 	<sos_export_package id="String"><br>
     * 		<sos_export_meta><br>
     * 			<table name="String"/><br>
     * 			<key_fields><br>
     * 				<field name="application" type="BLOB" len="100" scale="String"/><br>
     * 			</key_fields><br>
     * 			<fields><br>
     * 				<field name="application" type="BLOB" len="100" scale="String"/><br>
     * 			</fields><br>
     * 		</sos_export_meta><br>
     * 		<sos_export_data><br>
     * 			<sos_export_record name="String"><br>
     * 				<sos_export_fields><br>
     * 					<application>String</application><br>
     * 					<section>String</section><br>
     * 					<name>String</name><br>
     * 					<value>String</value><br>
     * 					<default_value>String</default_value><br>
     * 					<long_value>String</long_value><br>
     * 					<title>String</title><br>
     * 					<input_type>0</input_type><br>
     * 					<input_size>0</input_size><br>
     * 					<display_type>0</display_type><br>
     * 					<display_size>10</display_size><br>
     * 					<forced>0</forced><br>
     * 					<entry_type>0</entry_type><br>
     * 					<created>String</created><br>
     * 					<created_by>String</created_by><br>
     * 					<modified>String</modified><br>
     * 					<modified_by>String</modified_by><br>
     * 				</sos_export_fields><br>
     * 			</sos_export_record><br>
     * 		</sos_export_data><br>
     * 	</sos_export_package><br>
     * </sos_export><br><br>
     * 
     * 
     * 
     * @return boolean
     * @throws Exception
     */
    public boolean writeSettingFile() throws Exception {
        boolean retVal  =true;        
        try {
            
            sosLogger.info("Creating XML-Settings-File from config File " + configFile);  
            createWriterObject();           
            writeXML();
            output.close();
            sosLogger.info("XML-Settings-File is successfully created:" + configFile.replaceAll(".ini", ".xml"));
            
            return retVal;
        } catch (Exception e) {
            throw new Exception("\n ->..error in " + SOSClassUtil.getMethodName() + " " + e);
        }
    }
    
    /**
     * Schreibt den XML Datei.
     * 
     * @return
     * @throws Exception
     */
    private boolean writeXML() throws Exception {
        boolean retVal  =true;        
        String sectionname = "";
        Properties entrys = null;
        boolean buildAppName = false;
        try {
            
            if (output == null) {
                throw new Exception("..Bufferwriter Object ist null.");                
            }
            
            sosString = new SOSString();
            
            //Wenn die Applikationsname nicht bergeben wurde, dann wird diese hier 
            //in der Form application_sectionname gebildet.
            if (this.application.equals("application")) {
                buildAppName = true;
            }
            
			output.write("<?xml version=\"1.0\" encoding=\"" + getEncoding() +"\"?>");
			
			//<sos_export xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"	xsi:noNamespaceSchemaLocation="C:\eclipse\workspace\sos.tool\sos\settings\Untitled1.xsd" application="String" created="String"><br>
			 //<sos_export_package id="String"><br>
			
			output.write("<sos_export " + namespace + " application=\"" + application + "\" created=\"" + sos.util.SOSDate.getCurrentDateAsString() +"\">");

			output.write("<sos_export_package id=\""+ this.getTablename()+"\">");
			output.write("<sos_export_meta>");
				output.write("<table name=\""+ getTablename()+"\" />");
				
				output.write("<key_fields>");
					output.write("<field name=\"application\" type=\"VARCHAR\" len=\"100\" scale=\"\" />"); 
					output.write("<field name=\"section\" type=\"VARCHAR\" len=\"100\" scale=\"\" />"); 
					output.write("<field name=\"name\" type=\"VARCHAR\" len=\"100\" scale=\"\" />"); 
				output.write("</key_fields>");
				
				output.write("<fields>");
					output.write("<field name=\"application\" type=\"VARCHAR\" len=\"100\" scale=\"\" /> ");
					output.write("<field name=\"section\" type=\"VARCHAR\" len=\"100\" scale=\"\" /> ");
					output.write("<field name=\"name\" type=\"VARCHAR\" len=\"250\" scale=\"\" /> ");
					output.write("<field name=\"value\" type=\"VARCHAR\" len=\"1000\" scale=\"\" /> ");
					output.write("<field name=\"default_value\" type=\"VARCHAR\" len=\"1000\" scale=\"\" /> ");
					output.write("<field name=\"long_value\" type=\"BLOB\" len=\"4000\" scale=\"\" /> ");
					output.write("<field name=\"title\" type=\"VARCHAR\" len=\"100\" scale=\"\" /> ");
					output.write("<field name=\"input_type\" type=\"NUMBER\" len=\"22\" scale=\"\" /> ");
					output.write("<field name=\"input_size\" type=\"NUMBER\" len=\"22\" scale=\"\" /> ");
					output.write("<field name=\"display_type\" type=\"NUMBER\" len=\"22\" scale=\"\" /> ");
					output.write("<field name=\"display_size\" type=\"NUMBER\" len=\"22\" scale=\"\" /> ");
					output.write("<field name=\"forced\" type=\"NUMBER\" len=\"22\" scale=\"\" /> ");
					output.write("<field name=\"entry_type\" type=\"NUMBER\" len=\"22\" scale=\"\" /> ");
					output.write("<field name=\"created\" type=\"DATE\" len=\"7\" scale=\"\" /> ");
					output.write("<field name=\"created_by\" type=\"VARCHAR\" len=\"100\" scale=\"\" /> ");
					output.write("<field name=\"modified\" type=\"DATE\" len=\"7\" scale=\"\" /> ");
					output.write("<field name=\"modified_by\" type=\"VARCHAR\" len=\"100\" scale=\"\" /> ");
				output.write("</fields>");
				
				output.write("</sos_export_meta>");
			        		
			    output.write("<sos_export_data>");
			    for (int i = 0; i < this.sections.size(); i++) {			    
			        sectionname = sosString.parseToString(sections.get(i));
			        if (buildAppName) {
			            application = "application_" + sectionname;
			        }
			        
			        entrys = this.settings.getSection(sectionname);
			        Iterator keys = entrys.keySet().iterator();
			        Iterator values = entrys.values().iterator();
			        while(keys.hasNext()) {
			            output.write("<sos_export_record name=\"" + tablename +"\">");
				            output.write("<sos_export_fields>");
					            output.write("<application><![CDATA[" + application +"]]></application>");
					            output.write("<section><![CDATA[" + sectionname +"]]></section>");
					            output.write("<name><![CDATA[" + keys.next() +"]]></name>");
					            output.write("<value><![CDATA[" + values.next() +"]]></value>");
					            output.write("<default_value></default_value>");
					            output.write("<long_value></long_value>");
					            output.write("<title>"+sectionname+"</title>");
					            output.write("<input_type>0</input_type>");
					            output.write("<input_size>0</input_size>");
					            output.write("<display_type>0</display_type>");
					            output.write("<display_size>10</display_size>");
					            output.write("<forced>0</forced>");
					            output.write("<entry_type>0</entry_type>");
					            output.write("<created>"+SOSDate.getCurrentDateAsString() + "</created>");
					            output.write("<created_by>String</created_by>");
					            output.write("<modified>"+SOSDate.getCurrentDateAsString() +"</modified>");
					            output.write("<modified_by>String</modified_by>");
					        output.write("</sos_export_fields>");
			            output.write("</sos_export_record>");
			        }}
			        	
			   	output.write("</sos_export_data>");
			  output.write("</sos_export_package>");		
			output.write("</sos_export>");			
            return retVal;
        } catch (Exception e) {
            throw new Exception("\n ->..error in " + SOSClassUtil.getMethodName() + " " + e);
        }
    }
    
    

    /**
     * Erstellt das BufferedWriter Objekt. 
     * @return boolean
     * @throws Exception
     */
    private boolean createWriterObject() throws Exception {
        boolean retVal  =true;
        String fileName = "";
        try {
            fileName = this.configFile.replaceAll(".ini", ".xml");                       
            try {
                output = new BufferedWriter((new FileWriter(fileName)));
            } catch (IOException e) {
                throw new Exception(
                        "..error while creating Bufferwriter in WriteXML: " + e);
            }
            
            return retVal;
        } catch (Exception e) {
            throw new Exception("\n ->..error in " + SOSClassUtil.getMethodName() + " " + e);
        }
    }

    /**
     * Diese Methode liefert den Wert der Property <i>encoding</i>. 
     * Der Typ der Property <i>encoding</i> ist String.
     * 
     * @return String.
     */
    public String getEncoding() {
        return encoding;
    }
    /**
     * Hier wird ein Wert zu den Property <i>encoding</i> gesetzt.
     * Der Typ der Property <i>encoding</i> ist <i>String</i>.
     *
     * @param encoding
     */
    public void setEncoding(String encoding) {
        this.encoding = encoding;
    }
    /**
     * Hier wird ein Wert zu den Property <i>namespace</i> gesetzt.
     * Der Typ der Property <i>namespace</i> ist <i>String</i>.
     *
     * @param namespace
     */
    public void setNamespace(String namespace) {
        this.namespace = namespace;
    }
    /**
     * Hier wird ein Wert zu den Property <i>application</i> gesetzt.
     * Der Typ der Property <i>application</i> ist <i>String</i>.
     *
     * Die Profile Konfigurationsdatei hat keinen Applikation. Daher
     * wird die Applikationname hier gesetzt. Ansonst wird die applikationname
     * gebildet aus applikation_[sektionname], weil alle Eintrge zu einem 
     * Sektion auch zu einem Applikation gehren. 
     * 
     * @param application
     */
    public void setApplication(String application) {
        this.application = application;
    }
    /**
     * Hier wird ein Wert zu den Property <i>sosLogger</i> gesetzt.
     * Der Typ der Property <i>sosLogger</i> ist <i>SOSStandardLogger</i>.
     *
     * @see sos.util.SOSStandardLogger
     * @param sosLogger
     */
    public void setSosLogger(SOSStandardLogger sosLogger) {
        this.sosLogger = sosLogger;
    }
    /**
     * Diese Methode liefert den Wert der Property <i>configFile</i>. 
     * Der Typ der Property <i>SOSSettingsConverter</i> ist String.
     * 
     * @return String.
     */
    public String getConfigFile() {
        return configFile;
    }
    /**
     * Hier wird ein Wert zu den Property <i>configFile</i> gesetzt.
     * Der Typ der Property <i>configFile</i> ist <i>String</i>.
     *
     * @param configFile
     */
    public void setConfigFile(String configFile) throws Exception {
        try {
            this.configFile = configFile;
            this.isFile(configFile);            
        } catch (Exception e) {
            throw new Exception ("\n ->..error in " + SOSClassUtil.getMethodName() + " " + e);
        }
    }
    
    /**
     * berprft, ob die Konfigutaionsdatei wirklich existiert
     * @param configFile
     */
    public void isFile(String configFile) throws Exception {
        try {           
            if (!(new File(configFile).isFile())) {
                throw new Exception ("missing config File: " + configFile);
            }
        } catch (Exception e) {
            throw new Exception ("\n ->..error in " + SOSClassUtil.getMethodName() + " " + e);
        }
    }
    /**
     * Diese Methode liefert den Wert der Property <i>tablename</i>. 
     * Der Typ der Property <i>tablename</i> ist String.
     * 
     * @return String.
     */
    public String getTablename() {
        return tablename;
    }
    /**
     * Hier wird ein Wert zu den Property <i>tablename</i> gesetzt.
     * Der Typ der Property <i>tablename</i> ist <i>String</i>.
     *
     * @param tablename
     */
    public void setTablename(String tablename) {
        this.tablename = tablename;
    }
}
