package sos.settings.converter;

import java.util.*;

import sos.util.*;
import sos.settings.*;
import java.io.*;

/**
 * <p>Title: sos.settings.converter.SOSSettingsXML2ProfileConverter.java </p>
 * <p>Description: Ldt eine XML-Konfigurationsdatei und schreibt eine INI-Konfigurationsdatei.</p>
 * <p>Copyright: Copyright (c) 09.07.2004</p>
 * <p>Company: SOS GmbH</p>
 * @author Mrvet ksz
 * @resource sos.settings.jar; sos.util.jar, xerces.jar
 */
public class SOSSettingsXML2ProfileConverter extends SOSSettingsConverter {

    /**
     * Attribut: SOSStandardLogger Objekt.
     * @see sos.util.SOSStandardLogger 
     * */
    private SOSStandardLogger sosLogger = null;
    
    /** Attribut: Name der Konfigurationsdatei*/
    private String configFile = "";
        
    /** Attribut: Alle Eintrge der Sektion*/
    private ArrayList sections = null;
    
    /** Attribut: BufferWriter */
    private BufferedWriter output = null;
       
    /** Attribut: SOSString Objekt */
    private SOSString sosString = null;
    
    /** Attribut: SOSProfileSettings Object*/
    private SOSXMLSettings settings = null;
    
    /**
     * Konstruktor
     * @exception Exception
     */
    public SOSSettingsXML2ProfileConverter()throws Exception {
        super();        
    }

    /**
     * @param sosLogger_
     * @throws Exception
     */
    public SOSSettingsXML2ProfileConverter(SOSStandardLogger sosLogger_)
            throws Exception {
        super(sosLogger_);
    }
    
    /**
     * Konstruktor.
     * 
     * @param SOSStandardLogger
     * @see sos.util.SOSStandardLogger
     * @throws Exception
     */
    public SOSSettingsXML2ProfileConverter(String configFile_, SOSStandardLogger sosLogger_)
    		throws Exception {
        super(configFile_, sosLogger_);
        
    }
    
    /**
     * Ldt die XML-Konfigurationsdatei. <br>
     * Diese Methode liefert true, wenn die Konfigurationsdatei erfolgreich 
     * geladen wurde. 
     * 
     * @return boolean
     * @throws Exception
     */
    public boolean loadSettings() throws Exception {
        boolean retVal  =true;
        try {
            settings = new SOSXMLSettings(configFile, sosLogger);  
	        sections = settings.getListOfXMLFields();            
            return retVal;
        } catch (Exception e) {
            throw new Exception("\n ->..error in " + SOSClassUtil.getMethodName() + " " + e);
        }
    }
    
    /**
     * Schreibt die Konfigurationsdatei in eine neue Konfigurationsdatei. 
     * -> von XML zu INI Datei.
     * Wenn die INI-Konfigutaionsdatei erfolgreich geschrieben wurde, dann 
     * liefert die Methode true.
     * 
     * @return boolean
     * @throws Exception
     */
    public boolean writeSettingFile() throws Exception {
        boolean retVal  =true;
        String sectionname = "";
        Properties entrys = null;
        try {
            sosLogger.info("Creating INI-Settings-File from config File " + configFile);  
            createWriterObject();           
            writeProfile();
            output.close();
            sosLogger.info("INI-Settings-File is successfully created:" + configFile.replaceAll(".xml", ".ini"));                        
            return retVal;
        } catch (Exception e) {
            throw new Exception("\n ->..error in " + SOSClassUtil.getMethodName() + " " + e);
        }
    }
    /**
     * Schreibt die INI Datei.
     * 
     * 
     * @throws Exception
     */
    private void writeProfile() throws Exception {
                     
        boolean buildAppName 	= false;
        HashMap entrys 			= null;
        String sectionname 		= "";
        try {
            sosString = new SOSString();
            for (int i = 0; i < this.sections.size(); i++) { 			    
                entrys = (HashMap)sections.get(i);               
                if (sectionname == null || !sectionname.equalsIgnoreCase(sosString.parseToString(entrys.get("section")))) {
                    sectionname  = sosString.parseToString(entrys.get("section"));
                    output.write("\n\n[" + entrys.get("section") +"]");
                }
                
                output.write("\n; application=" + entrys.get("application"));                
                output.write("\n" + entrys.get("name") + "=" + entrys.get("value"));
                
            }                       
        } catch (Exception e) {
            throw new Exception("\n ->..error in " + SOSClassUtil.getMethodName() + " " + e);
        }
    }
    
    

    /**
     * Erstellt das BufferedWriter Objekt. 
     * 
     * @throws Exception
     */
    private void createWriterObject() throws Exception {       
        String fileName = "";
        try {
            fileName = this.configFile.replaceAll(".xml", ".ini");                       
            try {
                output = new BufferedWriter((new FileWriter(fileName)));
            } catch (IOException e) {
                throw new Exception(
                        "..error while creating Bufferwriter in WriteXML: " + e);
            }                        
        } catch (Exception e) {
            throw new Exception("\n ->..error in " + SOSClassUtil.getMethodName() + " " + e);
        }
    }
    /**
     * Diese Methode liefert den Wert der Property <i>configFile</i>. 
     * Der Typ der Property <i>configFile</i> ist String.
     * Die configfile ist dioe XML-Konfigurationsdatei, die in INI-Konfigurationsdatei
     * geschrieben werden soll.
     * 
     * @return String.
     */
    public String getConfigFile() {
        return configFile;
    }
    /**
     * Hier wird ein Wert zu den Property <i>configFile</i> gesetzt.
     * Der Typ der Property <i>configFile</i> ist <i>String</i>.
     *
     * @param configFile
     */
    public void setConfigFile(String configFile) {
        this.configFile = configFile;
    }
    
    /**
     * Hier wird ein Wert zu den Property <i>sosLogger</i> gesetzt.
     * Der Typ der Property <i>sosLogger</i> ist <i>SOSStandardLogger</i>.
     *
     * @see sos.util.SOSStandardLogger
     * @param sosLogger
     */
    public void setSosLogger(SOSStandardLogger sosLogger) {
        this.sosLogger = sosLogger;
    }
}
