// $Id: factory_rtf.h 3546 2005-04-03 08:30:29Z jz $

#ifndef __FACTORY_RTF_PARSER_H
#define __FACTORY_RTF_PARSER_H

#include "../kram/sosobj.h"
#include "../zschimmer/rtf/rtf.h"
#include "../zschimmer/rtf/rtf_doc.h"

namespace sos {
namespace factory {

using ::std::map;
using ::std::vector;
using ::std::stack;

//------------------------------------------------------------------------Rtf_constructor

struct Rtf_constructor : Ifactory_rtf_constructor,     // hostole.odl
                         Constructor,
                         zschimmer::rtf::Has_rtf_context
{
    typedef zschimmer::rtf::Entity                  Entity;
    typedef zschimmer::rtf::Simple_entity           Simple_entity;
    typedef zschimmer::rtf::Param_entity            Param_entity;
    typedef zschimmer::rtf::Destination_entity      Destination_entity;
    typedef zschimmer::rtf::Entity_ptr              Entity_ptr;
    typedef zschimmer::rtf::Entity_ptr_with_prop    Entity_ptr_with_prop;
    typedef zschimmer::rtf::Entity_listelem         Entity_listelem;
    typedef zschimmer::rtf::Doc                     Doc;

    struct Doc_entity
    { 
                                        Doc_entity      ( zschimmer::rtf::Has_rtf_context* c ) : _start(c), _end(c) {}

        int                            _doc_no;         // Index fr Rtf_constructor::_templates[]
        zschimmer::rtf::Simple_entity* _first;          // Start des (Sub-)Dokuments
        zschimmer::rtf::Entity_ptr     _start;                                      
        zschimmer::rtf::Entity_ptr     _end; 
        bool                           _in_script;      // In <%...%>
    };

    typedef map< int, Doc_entity >      Rtf_line_map;



    void*                       operator new            ( size_t size )                     { return sos_alloc( size, "Rtf_constructor" ); }
    void                        operator delete         ( void* ptr )                       { sos_free( ptr ); }

                                Rtf_constructor         ( Site*, Has_rtf_context* );
                               ~Rtf_constructor         ();

    USE_SOS_OLE_OBJECT

    // IFactory_global_object:
    STDMETHODIMP                Put                     ( VARIANT* v, VARIANT_BOOL raw )    { return Constructor::Put( v, raw ); }
    STDMETHODIMP                Put_bstr                ( BSTR bstr, VARIANT_BOOL raw )     { return Constructor::Put_bstr( bstr, raw ); }
    STDMETHODIMP                Put_formatted           ( VARIANT* v, Ausgabeformat f )     { return Constructor::Put_formatted( v, f ); }
    STDMETHODIMP                Embed                   ( BSTR name, Embed_flags flags )    { return Constructor::Embed( name, flags ); }
    
    STDMETHODIMP                Rtf_prop                ( int, int );
    STDMETHODIMP                Rtf_put                 ( int, int );
    STDMETHODIMP                Rtf_open                ( int );                            // {...
    STDMETHODIMP                Rtf_close               ();                                 // ...}
    STDMETHODIMP                Rtf_include_begin       ();                                 // ...}
    STDMETHODIMP                Rtf_include_end         ();                                 // ...}

    STDMETHODIMP                Rtf_start               ();
    STDMETHODIMP                Rtf_finish              ();
    STDMETHODIMP            put_Suppress_subdocument    ( VARIANT_BOOL b )                  { _suppress_subdocument = b != 0;  return S_OK; }
    STDMETHODIMP            get_Suppress_subdocument    ( VARIANT_BOOL* result )            { *result = _suppress_subdocument? VARIANT_TRUE : VARIANT_FALSE;  return S_OK; }

    void                        initialize              ();
    void                        close                   ();
    void                        close_doc_file          ();

    void                        put_string              ( const string&, bool raw );
    void                        rtf_prop                ( int prop, int param );
    void                        rtf_put                 ( int begin, int end );
    void                        rtf_open                ( int );
    void                        rtf_close               ();
    void                        start                   ();
    void                        finish                  ();
    void                        handle_inserted_entity  ( Simple_entity* );
    void                        set_error               ( const Xc& );
    string                      error_filename          ();
    string                      error_document          ();
    void                        close_output_file       ()                                  { close_doc_file(); }

    void                        map_next_lines          ( int doc_no, int lineno, Simple_entity* first, const Entity_ptr& start, const Entity_ptr& end, bool in_script );
    void                        map_next_line           ( int doc_no, int lineno, Simple_entity* first, const Entity_ptr& start, const Entity_ptr& end, bool in_script, const string& line_for_debug );
    Entity_listelem*            append_entity           ( const Simple_entity& );

    Fill_zero                  _zero_;
    vector<string>             _template_names;
    zschimmer::ptr<Doc>        _doc;
    zschimmer::rtf::Flags      _allowed_properties;
  //Simple_entity*             _first_entity_after_start; // vor start() und nach finish(): NULL. Falls finish() fehlt wird ab hier gelscht.
    vector< zschimmer::ptr<Doc> > _templates;           // Vom Rtf_parser, der vor dem Skriptlauf stirbt, eingelesene Dokumente, 
    int                        _error_doc;              // Index der fehlerhaften Vorlage (_templates) oder -1
    Rtf_line_map               _rtf_line_map;           // Skriptzeilennummer -> Entity*
    string                     _work_filename;
    z::Stream_file             _work_file;
  //fstream                    _work_file;
  //int                        _work_file_opened;
    int64                      _work_file_head_end;     // Position in _work_file, bei der der Dokumentkopf endet
    bool                       _suppress_subdocument;           // Wenn das laufende Subdokument nicht ausgegeben werden soll
    bool                       _last_subdocument_suppressed;    // Das letzte Subdokument ist nicht ausgegeben worden
    zschimmer::rtf::Properties _properties;
    int                        _include;
    bool                       _ignore_next_sectd;      // Fr OpenOffice nach #include erstes \sectd in der Vorlage unterdrcken
  //bool                       _sect_encountered;
    bool                       _titlepg_inserted;
};

//----------------------------------------------------------------------------Rtf_parser 

struct Rtf_parser : Parser,
                    zschimmer::rtf::Has_rtf_context
{
    BASE_CLASS( Parser )

    typedef zschimmer::rtf::Code                    Code;
    typedef zschimmer::rtf::Entity                  Entity;
    typedef zschimmer::rtf::Simple_entity           Simple_entity;
    typedef zschimmer::rtf::Param_entity            Param_entity;
    typedef zschimmer::rtf::Entity_listelem         Entity_listelem;
    typedef zschimmer::rtf::Entity_ptr              Entity_ptr;
    typedef zschimmer::rtf::Entity_ptr_with_prop    Entity_ptr_with_prop;
    typedef zschimmer::rtf::Properties              Properties;
    typedef zschimmer::rtf::Doc                     Doc;

    enum Symbol 
    {
        sym_none,
        sym_eof,
        sym_script,             // <% ... %>
        sym_preprocessor,       // <% #... %>
        sym_expr,               // &... oder &(...)
        sym_text,
    };

  //friend ostream&             operator <<             ( ostream&, Symbol );

    struct Token
    {
        Symbol                 _symbol;
        Source_pos             _pos;
        string                 _text;
        Ausgabeformat          _format;                 // Fr &$(...)
    };
    
  //friend ostream&             operator <<             ( ostream&, const Token& );

                                Rtf_parser              ( Has_rtf_context* );
                               ~Rtf_parser              ();

    void                        init                    ();
    void                        parse                   ( const char* script_text );
    void                        parse                   ( const string& script_text )               { parse( script_text.c_str() ); }
    void                        parse                   ();

    int                         eat_char                ();
    string                      eat_string              ();
    string                      eat_rtf_string          ();
    bool                        in_script_style         ();     // In einer Absatzformatvorlage fr Skript?
    void                        eat_token               ();
    void                        convert_quotes          ( string* );
    void                        put_props               ( const Properties& );
    void                        rtf_put                 ( Entity_listelem* start, Entity_listelem* end );
    void                        rtf_put_2               ( Entity_listelem* start, Entity_listelem* end );
    bool                        has_text                ( Simple_entity* );

    void                        merge                   ();
    void                        merge_table             ( Code table_code );
    void                        merge_listtable         ();
    void                        merge_listoverridetable ();
    void                        merge_colortbl          ();
  //void                        merge_properties        ( zschimmer::rtf::Flags );
    void                        modify_numbers          ( Entity_listelem*, bool but_lists = false );
    void                        map_merged_code         ( Code table_code, int number_in_template, int number_in_doc, const string& name = "" );
    void                        add_to_doc              ( Code property, Entity_listelem*, const string& debug_text = "" );
    void                        change_id               ( Simple_entity*, map<int,bool>* used_map, map<int,int>* m, int* id );
    map<int,int>&               number_map              ( Code table_code )  { return _number_map[ _rtf_constructor->_doc->_header.table_index(table_code) ]; }

    Bool                       _obj_is_type             ( Sos_type_code t ) const               { return t == tc_factory_Rtf_parser || Base_class::_obj_is_type( t ); }


    Fill_zero                  _zero_;

    zschimmer::ptr<Doc>        _template;               // Einzulesendes Dokument (Hauptdokumentvorlage, #include oder embed)
    int                        _template_index;         // _rtf_constructor->_templates[_template_index] == _template
    zschimmer::ptr<Doc>        _main_doc;               // Bei #include _main_template, bei embed _rtf_construction->_doc, Ziel fr merge()
    Rtf_constructor*           _rtf_constructor;
    bool                       _dont_merge;             // Bei RTF-Variable, die schon alle Nummern (\fN etc.) korrekt enthlt
  //bool                       _convert_doc_to_section_properties;  // Bei _different_templates_allowed: Ab zweiten Dokument Dokument- zu Abschnittseigenschaften
    Simple_entity*             _first_entity;
    Entity_ptr_with_prop       _entity_ptr;
    Entity_ptr                 _line_entity_ptr;        // Anfang der Skriptzeile
    Entity_ptr_with_prop       _text_start;
    Entity_ptr                 _script_end;             // Zeigt bei sym_script auf "%>"
    Simple_entity*             _last_put_props;
    int                        _char;
    Source_pos                 _char_pos;
    Source_pos                 _eaten_char_pos;
    Token                      _token;
    bool                       _in_script;              // In <% ... %>
    bool                       _in_script_style;
    int                        _script_para_style_no;   // Nummer der Absatzformatvorlage fr Skript
    int                        _script_char_style_no;   // Nummer der Zeichenformatvorlage fr Skript
    bool                       _has_script;             // Im RTF-Text sind Skriptanweisungen oder Ausdrcke vorgekommen
    Simple_entity*             _debug_outer_entity;

    map<int,int>               _number_map[ zschimmer::rtf::doc_table_count ];
    map<int,int>               _listid_map;
    map<int,int>               _listtemplateid_map;
    map<int,int>               _leveltemplateid_map;

    map<int,bool>              _used_listid_map;
    map<int,bool>              _used_listtemplateid_map;
    map<int,bool>              _used_leveltemplateid_map;
};


} //namespace factory
} //namespace sos


#endif
