// $Id: system_command.cxx 3546 2005-04-03 08:30:29Z jz $
//  2004 Joacim Zschimmer

#include "zschimmer.h"
#include "log.h"
#include "system_command.h"
#include "file.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/timeb.h>

#ifdef Z_WINDOWS
#   include <windows.h>
#   include <io.h>
#else
#   include <unistd.h>
#   include <locale.h>
#   include <sys/time.h>
#   include <signal.h>
#endif

using namespace std;


namespace zschimmer {

//--------------------------------------------------------------------------System_command::execute

void System_command::execute( const string& cmd )
{
    File stderr_file;
    File stdout_file;
    string my_cmd = cmd;

    _stderr_text = "";

    if( cmd.find( "2>" ) == string::npos )
    {
        stderr_file.create_temporary( File::open_unlink_later );
        my_cmd.append( " 2>" + stderr_file.filename() );
    }

    
    if( !has_regex( cmd, "[^2]>" ) )
    {
        stdout_file.create_temporary( File::open_unlink_later );
        my_cmd.append( " >" + stdout_file.filename() );
    }


    Z_LOG( "system " << my_cmd << "\n" );
    int ret = system( my_cmd.c_str() );
    int system_errno = errno;

#   ifdef Z_WINDOWS
        _exit_code = ret;
#    else
        _exit_code = ret == -1? -1 : WEXITSTATUS(ret);
#   endif


    if( stderr_file.filename() != "" )
    {
        try {
            _stderr_text = string_from_file( stderr_file.filename() );
        }
        catch( const exception& ) {}

        //unlink( stderr_file.filename().c_str() );
    }

    if( stdout_file.filename() != "" )
    {
        try {
            _stdout_text = string_from_file( stdout_file.filename() );
        }
        catch( const exception& ) {}

        //unlink( stdout_file.filename().c_str() );
    }

    try
    {
        if( ret == -1 )  throw_errno( system_errno, "system" );

#       ifdef Z_WINDOWS
            if( ret != 0 )  throw_xc( "Z-4005", as_string(ret), cmd );
#        else
            if( WIFSIGNALED(ret) )                      throw_xc( "Z-4006", as_string(WTERMSIG(ret)), cmd );
            if( WIFEXITED(ret)  &&  WEXITSTATUS(ret) )  throw_xc( "Z-4005", as_string(WEXITSTATUS(ret)), cmd );
#       endif
    }
    catch( const Xc& x )
    {
        _error = true;
        if( _throw_xc )  throw;
        _xc = x;
    }
}

//-------------------------------------------------------------------------------------------------

} //namespace zschimmer
