// $Id: xml_libxml2.cxx 3727 2005-06-21 16:55:46Z jz $

// Dokumentation von libxml2 in http://xmlsoft.org/html/libxml-lib.html

#include "zschimmer.h"
#include "xslt_libxslt.h"
#include "log.h"

#include "../3rd_party/libxml2/libxml.h"
#include "../3rd_party/libxml2/include/libxml/xmlmemory.h"
#include "../3rd_party/libxml2/include/libxml/xmlIO.h"

#include "../3rd_party/libxslt/libxslt/xslt.h"
#include "../3rd_party/libxslt/libxslt/xsltInternals.h"
#include "../3rd_party/libxslt/libxslt/transform.h"
#include "../3rd_party/libxslt/libxslt/xsltutils.h"



using namespace std;
using namespace zschimmer::com;


namespace zschimmer {
namespace xml_libxml2 {
namespace xml {

//-------------------------------------------------------------------------Libxml_output_buffer_ptr
/*
struct Libxml_output_buffer_ptr
{
    Libxml_output_buffer_ptr() 
    : 
        _ptr(NULL) 
    {
    }


    ~Libxml_output_buffer_ptr()
    {
        if( _ptr )  xmlOutputBufferClose( _ptr );
    }


    operator xmlOutputBufferPtr () 
    { 
        return _ptr; 
    }


    xmlOutputBufferPtr         _ptr;
};
*/
//-------------------------------------------------------------------------------------------------

struct Xslt_parameters
{
    Xslt_parameters( int count )
    :
        _n(count),
        _array( new const char*[ 2*count + 1 ] )
    {
        for( int i = 0; i < 2*count + 1; i++ )  _array[i] = NULL;
    }


    ~Xslt_parameters()
    {
        if( _array )
        {
            for( int i = 0; i < 2*_n + 1; i++ )  delete[] _array[i];
            delete[] _array;
        }
    }


    void set( int i, const string& name, const string& value )
    {
        assert( i >= 0  &&  i < _n );

        delete[] _array[ i   ];   _array[ i   ] = strdup( name.c_str() );
        delete[] _array[ i+1 ];   _array[ i+1 ] = strdup( value.c_str() );
    }

    int                        _n;
    const char**               _array;
};

//--------------------------------------------------------------------------------------------const

static Error_code_text error_codes[] =
{
    { "LIBXSLT-001", "Kein Stylesheet geladen" },
    { "LIBXSLT-002", "Parametersyntax NAME=WERT nicht eingehalten: $1" },
    { "LIBXSLT-003", "Stylesheet nicht ladbar" },
    { "LIBXSLT-004", "Fehler bei der Anwendung des Stylesheets auf das XML-Dokument" },
    { NULL         , NULL }
};

//-------------------------------------------------------------------------------------------Z_INIT

Z_INIT( xml_libxslt )
{
    add_error_code_texts( error_codes );
}


//xsltSetTransformErrorFunc( 

//----------------------------------------------------------slt_stylesheet_ptr::Xslt_stylesheet_ptr

Xslt_stylesheet_ptr::Xslt_stylesheet_ptr( _xsltStylesheet* ptr )
:
    _ptr( ptr )
{
}

//---------------------------------------------------------slt_stylesheet_ptr::~Xslt_stylesheet_ptr

Xslt_stylesheet_ptr::~Xslt_stylesheet_ptr()
{
    if( _ptr )
    {
        xsltFreeStylesheet( _ptr );

        //? xsltCleanupGlobals();
	//? xmlCleanupParser();
    }
}

//-----------------------------------------------------------------Xslt_stylesheet::Xslt_stylesheet

Xslt_stylesheet::Xslt_stylesheet( Xslt_stylesheet_ptr* stylesheet_ptr )
: 
    _stylesheet_ptr(stylesheet_ptr)
{
}

//-----------------------------------------------------------------Xslt_stylesheet::Xslt_stylesheet

Xslt_stylesheet::Xslt_stylesheet( const string& xml_or_filename )
{
    if( is_xml( xml_or_filename ) )
    {
        load( Document_ptr( xml_or_filename ) );
    }
    else
    {
        load_file( xml_or_filename );
    }
}

//-----------------------------------------------------------------Xslt_stylesheet::Xslt_stylesheet

Xslt_stylesheet::Xslt_stylesheet( const BSTR xml_or_filename_bstr )
{
    if( is_xml( xml_or_filename_bstr ) )
    {
        load( Document_ptr( xml_or_filename_bstr ) );
    }
    else
    {
        load_file( string_from_bstr( xml_or_filename_bstr ) );
    }
}

//----------------------------------------------------------------Xslt_stylesheet::~Xslt_stylesheet

Xslt_stylesheet::~Xslt_stylesheet()
{
    release();
}

//-------------------------------------------------------------------------Xslt_stylesheet::release

void Xslt_stylesheet::release()
{
    _stylesheet_ptr = NULL;
}

//--------------------------------------------------------------------------Xslt_stylesheet::is_xml

bool Xslt_stylesheet::is_xml( const string& str )
{
    return string_begins_with( str, "<" );
}

//--------------------------------------------------------------------------Xslt_stylesheet::is_xml

bool Xslt_stylesheet::is_xml( const BSTR bstr )
{
    return bstr && bstr[0] == '<';
}

//-----------------------------------------------------------------Xslt_stylesheet::prepare_parsing

void Xslt_stylesheet::prepare_parsing()
{
    xmlSubstituteEntitiesDefault( 1 );
    xmlLoadExtDtdDefaultValue = 1;
}

//-----------------------------------------------------------------------Xslt_stylesheet::load_file

void Xslt_stylesheet::load_file( const string& path )
{
    _stylesheet_ptr = NULL;

    _path = path;   // Auch im Fehlerfall

    prepare_parsing();
    Activate_error_text activate_error_text ( this );

    _xsltStylesheet* p = xsltParseStylesheetFile( Utf8_string( path ).utf8() );
    if( !p )  throw_xc( "LIBXSLT-003", "xsltParseStylesheetFile", _error_text );

    _stylesheet_ptr = Z_NEW( Xslt_stylesheet_ptr( p ) );
}

//------------------------------------------------------------------------Xslt_stylesheet::load_xml

void Xslt_stylesheet::load( const Document_ptr& stylesheet )
{
    _stylesheet_ptr = NULL;
    _path = "";

    prepare_parsing();
    Activate_error_text activate_error_text ( this );

    _xsltStylesheet* p = xsltParseStylesheetDoc( stylesheet.ptr() );
    if( !p )  throw_xc( "LIBXSLT-003", "xsltParseStylesheetDoc", _error_text );

    _stylesheet_ptr = Z_NEW( Xslt_stylesheet_ptr( p ) );
}

//---------------------------------------------------------------------------Xslt_stylesheet::apply

Document_ptr Xslt_stylesheet::apply( const Document_ptr& document, const vector<string>& parameters )
{
    if( !_stylesheet_ptr )  throw_xc( "LIBXSLT-001" );

    Document_ptr    result;
    Xslt_parameters parameter_array ( parameters.size() );

    for( uint i = 0; i < parameters.size(); i++ )
    {
        const string& parameter = parameters[i];
        size_t equal = parameter.find( '=' );
        if( equal == string::npos )  throw_xc( "LIBXSLT-002", parameter );
        parameter_array.set( i, parameter.substr( 0, equal ), parameter.substr( equal + 1 ) );
    }

    Activate_error_text activate_error_text ( this );

    Z_LOG( "xsltApplyStylesheet()\n" );
    result = xsltApplyStylesheet( _stylesheet_ptr->_ptr, document.ptr(), parameter_array._array );
    if( !result )  throw_xc( "LIBXSLT-004", "xsltApplyStylesheet", _error_text );

    return result;
}

//------------------------------------------------------------Xslt_stylesheet::write_result_to_file

void Xslt_stylesheet::write_result_to_file( const Document_ptr& result_document, const string& filename )
{
  //xmlIndentTreeOutput = 1;

    int compression = 0;
    xsltSaveResultToFilename( filename.c_str(), result_document.ptr(), _stylesheet_ptr->_ptr, compression );

    /*
    File file ( filename, "w" );
    xsltSaveResultToFd( file, result_document._ptr, _ptr );
    file.release();
    */
}

//-----------------------------------------------------------------Xslt_stylesheet::xml_from_result
/*
string Xslt_stylesheet::xml_from_result( const Document_ptr& result_document )
{
    string                  result;
    Libxml_output_buffer_ptr    buffer;


    int byte_count = xsltSaveResultTo( buffer, result_document.ptr(), _ptr );
    if( byte_count < 0 )  throw_xc( "LIBXSLT-xsltSaveResultToString" );

    return sd( buffer );
}
*/
//-------------------------------------------------------------------------------------------------

} //namespace xml
} //namespace xml_libxml2
} //namespace zschimmer

//-------------------------------------------------------------------------------------------------
